import requests
from io import BytesIO, StringIO
from typing import Union
from pathlib import Path
from uquake.core.logging import logger


def get_test_picker_model(output_path: Union[str, Path] = 'test_picker_model.pickle'):
    result = requests.get('https://www.dropbox.com/s/7q2l2uo047rdm73/'
                          'picker_model.pickle?dl=1')
    content = BytesIO(result.content)
    with open(output_path, 'wb') as f_out:
        f_out.write(content.getbuffer())

def get_test_data(output_folder: Union[str, Path]):

    if isinstance(output_folder, str):
        output_folder = Path(output_folder)

    get_context_trace(output_folder)
    get_fixed_length_trace(output_folder)
    get_variable_length_trace(output_folder)
    get_catalog(output_folder)


def get_context_trace(output_folder: Union[str, Path]):
    url = 'https://www.dropbox.com/s/k9xg3cxk7l1f2i7/' \
          '5fc90cac82f5d31fc2acb7290b87f411.context_mseed?dl=1'

    logger.info('extracting context trace')

    result = requests.get(url)
    content = BytesIO(result.content)
    with open(output_folder / f'context.mseed', 'wb') as f_out:
        f_out.write(content.getbuffer())


def get_fixed_length_trace(output_folder: Union[str, Path]):
    url = 'https://www.dropbox.com/s/3co5it512oqu8cy/' \
          '5fc90cac82f5d31fc2acb7290b87f411.mseed?dl=1'

    logger.info('extracting fixed length trace')

    result = requests.get(url)
    content = BytesIO(result.content)
    with open(output_folder / f'fixed_length.mseed', 'wb') as f_out:
        f_out.write(content.getbuffer())


def get_variable_length_trace(output_folder: Union[str, Path]):
    url = 'https://www.dropbox.com/s/dhuq15i5vtvtee8/' \
          '5fc90cac82f5d31fc2acb7290b87f411.'

    logger.info('extracting variable length trace')

    result = requests.get(url)
    content = BytesIO(result.content)
    with open(output_folder / f'variable_length.mseed', 'wb') as f_out:
        f_out.write(content.getbuffer())


def get_catalog(output_folder: Union[str, Path]):
    url = 'https://www.dropbox.com/s/gsh15umzs7r6gaj/' \
          '5fc90cac82f5d31fc2acb7290b87f411.xml?dl=1'

    logger.info('extracting catalogue')

    result = requests.get(url)
    content = BytesIO(result.content)
    with open(output_folder / f'catalogue.xml', 'wb') as f_out:
        f_out.write(content.getbuffer())







# repo = Repo()
# git_url = 'https://jeanphilippemercier/useis_test_data.git'
#
# repo.clone_from(git_url, '.')


#
# @Author : Jean-Pascal Mercier <jean-pascal.mercier@agsis.com>
#
#
#
__doc__ = \
"""
TODO : BIG DESCRIPTION OF EXCHANGE DATA
"""

import numpy as np
import pickle
import copy
from functools import wraps


def memoize(fct):
    """
    This is a decorator which cache the result of a function based on the
    given parameter.
    """
    return_dict = {}

    @wraps(fct)
    def wrapper(*args, **kwargs):
        if args not in return_dict:
            return_dict[args] = fct(*args, **kwargs)
        return return_dict[args]
    return wrapper

# This is the data description for the input array describing the event
#
ev_dtype = [('name',       'str'),
            ('id',          'int'),
            ('position',    'float',    (3,)),
            ('delta_t',     'float')]

st_dtype = ev_dtype

tt_dtype = [('id',          'int'),
            ('event_id',    'int'),
            ('traveltime',  'float')]


class EKTTTable(object):
    """
    This object represent a
    :param data: u
    :param staid: u
    :param evnfile: u
    :param stafile: u
    """
    dtype = tt_dtype

    def __init__(self, data, staid, evnfile=None, stafile=None):
        try:
            for tname, ttype in tt_dtype:
                data[tname]
            import sys
            sys.stderr.write(str(data.size))
            sys.stderr.flush()
            data = np.array(data.__getitem__([tname for tname, ttype in 
                                              tt_dtype]), dtype = self.dtype)
        except ValueError as e:
            data = np.asarray(data, dtype = self.dtype)

        self.data               = data
        self.data.dtype         = self.dtype
        self.station_id         = staid

        self.__evn_file__   = evnfile
        self.__sta_file__   = stafile

    def __get_event_rows__(self):
        return self.event_table.data[self.data['event_id']]
    event_rows = property(__get_event_rows__)

    @memoize
    def __get_event_table__(self):
        return pickle.load(open(self.__evn_file__))
    event_table = property(__get_event_table__)

    def __get_station_row__(self):
        return self.station_table.data[self.station_id]
    station_row = property(__get_station_row__)

    @memoize
    def __get_station_table__(self):
        return pickle.load(open(self.__sta_file__))
    station_table = property(__get_station_table__)


class EKPunctualData(object):
    dtype = None

    def __init__(self, data, origin=None, scale=1):
        try:
            for tname, ttype in self.dtype:
                data[tname]
            self.data = data
        except ValueError as e:
            self.data = np.asarray(data, dtype=self.dtype)

        self.origin = tuple([0] * data['position'].shape[-1]) \
            if origin is None else origin
        self.scale = scale

    def __get_position_zyx__(self):
        return self.data['position']

    def __set_position_zyx__(self, pos):
        self.data['position'] = pos
    position_zyx = property(__get_position_zyx__, __set_position_zyx__)

    def __get_position_xyz__(self):
        position_zyx = self.data['position']
        position_xyz = np.empty_like(position_zyx)
        for i in range(position_zyx.shape[1]):
            position_xyz[:, i] = position_zyx[:, -(i + 1)]
        return position_xyz

    position_xyz = property(__get_position_xyz__)

    def __get_size__(self):
        return self.data.size
    size = property(__get_size__)

    def add_gaussian_noise(self, position=0, time=0):
        pass


class EKEventTable(EKPunctualData):
    dtype = ev_dtype


class EKStationTable(EKPunctualData):
    dtype = st_dtype


class EKImageData(object):
    """
    This object represent a geometric structure that representing a regular
    array of points positionned in space.

    :param shape_or_data: The numpy array or the shape of the underlying data
    :param spacing: The spacing of the grid
    :param origin: A Tuple representing the position of the lower left corner \
            of the grid
    """
    def __init__(self, shape_or_data, spacing = 1, origin = None):
        if isinstance(shape_or_data, np.ndarray):
            self.data = shape_or_data
        else:
            self.data = np.zeros(shape_or_data)

        self.origin = tuple([0] * self.data.ndim) if origin is None else origin
        self.spacing = spacing

    def transform_to(self, values):
        return (values - self.origin) / self.spacing

    def transform_from(self, values):
        return (values + self.origin) * self.spacing

    def check_compatibility(self, other):
        return (self.shape == other.shape) and \
                (self.spacing == other.spacing) and \
                (self.origin == other.origin)

    def __get_shape__(self):
        return self.data.shape
    shape = property(__get_shape__)

    def homogenous_like(self, value):
        data = np.empty_like(self.data)
        data.fill(value)
        return EKImageData(data, self.spacing, origin=self.origin)

    def copy(self):
        cp = copy.deepcopy(self)
        return cp

    def SaveNLL(self):
        pass

    def LoadNLL(self):
        pass

    ### ADDING method to save and load on data from/to NLL







import numpy as np
import pandas as pd
from ..core.project_manager import ProjectManager
from pathlib import Path
import os
import shutil
from ..settings.settings import Settings
from uquake.waveform.pick import snr_ensemble_re_picker, calculate_snr
from uquake.core.event import (Pick, WaveformStreamID, ResourceIdentifier,
                               Arrival, Origin, Event)
from uquake.core import UTCDateTime
import uquake
from ..ai.model import AIPicker
from datetime import datetime
from uquake.core.stream import Stream
from uquake.core.inventory import Inventory
from typing import List
from obspy.realtime.signal import kurtosis
from numpy.fft import fftshift
from loguru import logger
from typing import List


class PickerResult(object):
    def __init__(self, new_picks, initial_picks, snr_threshold, stream):
        self.new_picks = new_picks
        self.initial_picks = initial_picks
        self.snr_threshold = snr_threshold
        self.stream = stream

        for k, pick in enumerate(self.new_picks):
            network = pick.waveform_id.network_code
            station = pick.waveform_id.station_code
            location = pick.waveform_id.location_code

            snr = calculate_snr(self.stream.select(network=network,
                                                   station=station,
                                                   location=location
                                                   )[0],
                                pick.time,
                                pre_wl=0.05, post_wl=0.02)

            self.new_picks[k].snr = snr

    def export_event(self):
        return self.append_event()

    @property
    def picks(self):
        picks = []
        for pick in self.new_picks:
            if self.snr_threshold is not None:
                if pick.snr < self.snr_threshold:
                    continue
            picks.append(pick)

        return picks

    def append_event(self, event: Event = None):

        snr_threshold = self.snr_threshold
        arrivals = []
        picks = []

        for pick in self.new_picks:
            picks.append(pick)

        for pick in self.initial_picks:
            picks.append(pick)

            arrival = Arrival(pick_id=pick.resource_id,
                              phase=pick.phase_hint)
            arrivals.append(arrival)

            origin = Origin(arrivals=arrivals)

        if event is None:
            event = Event(picks=picks, origins=[origin])

            return event

        if isinstance(event, uquake.core.event.Catalog):
            event[0].origins.append(origin)
            for pick in self.new_picks:
                event[0].picks.append(pick)

            return event

        else:
            event.origins.append(origin)
            for pick in self.new_picks:
                event.picks.append(pick)

            return event

    # def measure_incidence(self, stream: Stream,
    #                       inventory: Inventory):
    #
    #     pass

    def write_simple_pick_file(self, filename):
        with open(filename, 'w') as pick_file_out:
            for pick in self.picks:
                network = pick.waveform_id.network_code
                station = pick.waveform_id.station_code
                location = pick.waveform_id.location_code
                channel = None
                if pick.waveform_id.channel_code is not None:
                    channel = pick.waveform_id.channel_code[0:2]

                if self.snr_threshold is not None:
                    if pick.snr < self.snr_threshold:
                        continue

                station_string = f'{network}.{station}.'
                # {location}.{channel}*'
                line_out = f'{station_string},{pick.phase_hint.upper()},' \
                           f'{pick.time.timestamp}\n'
                pick_file_out.write(line_out)


class Picker(ProjectManager):

    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str):
        """
        Object to manage repicking within a project
        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        """

        super().__init__(base_projects_path, project_name, network_code)
        self.files.picker_settings = self.paths.config / 'picker_settings.toml'

        if not self.files.picker_settings.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                            '../settings/picker_settings_template.toml'

            shutil.copyfile(settings_template,
                            self.files.picker_settings)
            super().__init__(base_projects_path, project_name, network_code)

        self.files.picker_model = self.paths.ai_models / 'picker.pickle'

        self.ai_picker = None

        if self.files.picker_model.is_file():
            if self.files.picker_model.is_file():
                self.ai_picker = AIPicker.read(
                    self.files.picker_model)

    @staticmethod
    def read_nmx_pick(pick_file: str):

        nmx_picks = pd.read_csv(pick_file,
                                names=['site_id', 'phase', 'timestamp'])

        nmx_picks['datetime'] = [datetime.utcfromtimestamp(
            pick[1]['timestamp']) for pick in nmx_picks.iterrows()]
        out_picks = []
        for pick in nmx_picks.iterrows():
            pick = pick[1]
            if len(pick.site_id.split('.')) == 3:
                network, station, location = pick.site_id.split('.')
                channel = None
            elif len(pick.site_id.split('.')) == 4:
                network, station, location, channel = pick.site_id.split('.')

            waveform_id = WaveformStreamID(network_code=network,
                                           station_code=station,
                                           location_code=location,
                                           channel_code=channel)

            out_picks.append(
                Pick(waveform_id=waveform_id, phase_hint=pick.phase,
                     time=UTCDateTime(pick.datetime),
                     evaluation_mode='automatic',
                     evaluation_status='preliminary',
                     method_id=ResourceIdentifier('beamforming'),
                     method='beamforming'))

        return out_picks

    def add_picker_model_from_file(self, file_path):
        # shutil.copyfile(file_path, self.files.picker_model)
        self.ai_picker = AIPicker.read(file_path)
        self.ai_picker.write(self.files.picker_model)

    def add_picker_model(self, model):
        self.ai_picker = model
        self.ai_picker.write(self.files.picker_model)

    def synthetic_pick(self, origin_location, origin_time):
        tts = self.travel_times.travel_time(origin_location)

        picks = []
        for phase in ['P', 'S']:
            for tt in tts[phase]:
                t = origin_time + tt
                waveform_id = WaveformStreamID(network_code=self.network_code,
                                               station_code=self.station_code)
                # pk = Pick(time=t, )
                picks.append()



    def snr_repick(self, stream: uquake.core.stream.Stream,
                   picks: List[uquake.core.event.Pick],
                   ai_enhanced=False,
                   setting_section='snr_repicker'):

        """
        Repick using a snr based ensemble picker
        :param stream: stream containing the waveforms
        :type stream: :class: uquake.core.stream.Stream
        :param picks: a list of picks associated to the waveforms
        :type picks: a list of :class: uquake.core.event.Pick
        :return:
        """

        start_search_window = self.settings[
            setting_section].start_search_window
        end_search_window = self.settings[setting_section].end_search_window
        start_refined_search_window = \
            self.settings[setting_section].start_refined_search_window
        end_refined_search_window = \
            self.settings[setting_section].end_refined_search_window
        search_resolution = self.settings[setting_section].search_resolution
        pre_pick_window_len = self.settings[
            setting_section].pre_pick_window_len
        post_pick_window_len = self.settings[
            setting_section].post_pick_window_len

        snrs, new_picks = \
            snr_ensemble_re_picker(stream, picks,
                                   start_search_window=start_search_window,
                                   end_search_window=end_search_window,
                                   start_refined_search_window=
                                   start_refined_search_window,
                                   end_refined_search_window=
                                   end_refined_search_window,
                                   refined_window_search_resolution=
                                   search_resolution,
                                   snr_calc_pre_pick_window_len
                                   =pre_pick_window_len,
                                   snr_calc_post_pick_window_len
                                   =post_pick_window_len)

        out_picks = []
        for _i, new_pick in enumerate(new_picks):
            network = new_pick.waveform_id.network_code
            station = new_pick.waveform_id.station_code
            location = new_pick.waveform_id.location_code
            channel = new_pick.waveform_id.channel_code

            pre_wl = self.settings.snr_repicker.snr_pre_pick_window
            post_wl = self.settings.snr_repicker.snr_post_pick_window

            snr = calculate_snr(stream.select(network=network, station=station,
                                              location=location)[0],
                                new_pick.time, pre_wl=pre_wl, post_wl=post_wl)

            # if snr < self.settings.snr_repicker.snr_threshold:
            #     continue

            new_pick.snr = snr
            out_picks.append(new_pick)
            
        if ai_enhanced:
            if self.ai_picker is None:
                logger.error('no picker picker model in the project\n'
                             'to add a picker picker model please use the'
                             'add_ai_picker_from_file method')
            else:
                self.ai_pick(stream.copy(), out_picks)

        return PickerResult(new_picks=out_picks,
                            initial_picks=picks,
                            snr_threshold=
                            self.settings.snr_repicker.snr_threshold,
                            stream=stream)

    @staticmethod
    def origin_time_correction_from_picks(stream: Stream, picks: List[Pick],
                                          search_window: float = 0.2,
                                          kurtosis_window: float = 0.02,
                                          use_p: bool = True,
                                          use_s: bool = True,
                                          snr_threshold: float = None)\
            -> PickerResult:
        """
        estimate the origin time using the waveform and a list of picks.
        This function is used to correct the origin time obtained from
        beam forming.
        :param stream: the waveforms
        :type stream: uquake.core.stream.Stream
        :param picks: a list of picks
        :type picks: List[Pick]
        :param search_window: search window in seconds
        :type search_window: float
        :param kurtosis_window: length of the window used to calculate the
        Kurtosis function. see obspy.realtime.signal import kurtosis
        :type kurtosis_window: float
        :param use_p: use p picks
        :type use_p: bool
        :param use_s: use s picks
        :type use_s: bool
        :param snr_threshold: snr threshold for picks
        :type snr_threshold: float
        :return: time correction in second
        :rtype: uquake.core.UTCDateTime
        """

        wf = stream
        min_time = np.min([tr.stats.starttime for tr in wf])

        sampling_rate = wf[0].stats.sampling_rate
        window_length = int(search_window * sampling_rate)

        # stack data
        stacked_data = np.zeros(len(wf[0].data))
        for pick in picks:
            if (pick.phase_hint == 'P') and not use_p:
                continue
            if (pick.phase_hint == 'S') and not use_s:
                continue
            station = pick.waveform_id.station_code
            wf = wf.resample(sampling_rate)
            for tr in wf.select(station=station):
                if (tr.stats.starttime > pick.time) or \
                        (pick.time > tr.stats.endtime):
                    continue
                data = tr.data.astype(np.float32) ** 2
                data /= np.std(data)
                n_sample = int((min_time - pick.time) * tr.stats.sampling_rate)
                stacked_data += fftshift(np.roll(data, n_sample))

        stacked_trace = tr.copy()
        stacked_trace.data = stacked_data

        i_max = np.argmax(stacked_trace)

        k = kurtosis(stacked_trace, win=kurtosis_window)
        diff_k = np.diff(k)
        i_0 = int(i_max - window_length / 2)
        if i_0 < 0:
            i_0 = 0
        i_1 = int(i_max + window_length / 2)
        if i_1 > len(stacked_trace):
            i_1 = len(stacked_trace)
        o_i = np.argmax(np.abs(diff_k[i_0: i_1])) + i_0
        origin_time_correction = (o_i - len(stacked_trace) / 2) / sampling_rate

        out_picks = []
        for pick in picks:
            pick.time += origin_time_correction
            out_picks.append(pick)

        return PickerResult(out_picks, picks, snr_threshold, stream)

    def ai_pick(self, st: Stream, picks: list):
        import matplotlib.pyplot as plt

        st2 = st.copy()
        st2 = st2.detrend('demean').detrend('linear').taper(max_length=0.01,
                                                            max_percentage=1)

        st2 = st2.resample(sampling_rate=self.settings.ai_picker.sampling_rate)

        picks_out = []
        for pick in picks:
            out_pick = pick.copy()
            network = pick.waveform_id.network_code
            station = pick.waveform_id.station_code
            location = pick.waveform_id.location_code
            phase = pick.phase_hint

            pick_time = pick.time

            st3 = st2.copy().select(network=network, station=station,
                                    location=location)

            tr = st2.copy().select(network=network, station=station,
                                   location=location).composite()[0]

            predicted_time = self.ai_picker.predict_trace(tr.copy(),
                                                          pick_time,
                                                          phase)

            pick.time = predicted_time
            picks_out.append(pick)

        return picks_out


def measure_linearity_planarity(st: Stream, pick: Pick, window_length: float):
    network = pick.waveform_id.network_code
    station = pick.waveform_id.station_code
    location = pick.waveform_id.location_code

    st2 = st.select(network=network, station=station, location=location)

    wave_mat = []
    for tr in st2:
        window_start_sample = int((pick.time - tr.stats.starttime)
                                  * tr.stats.sampling_rate)
        window_end_sample = int(window_start_sample +
                                window_length * tr.stats.sampling_rate)

        wave_mat.append(tr.data[window_start_sample: window_end_sample])

    cov_mat = np.cov(np.array(wave_mat))

    eig_vals, eig_vects = np.linalg.eig(cov_mat)
    vals = np.sort(eig_vals)

    if pick.phase_hint == 'P':
        out_value = 1 - np.linalg.norm([vals[0], vals[1]]) / vals[2]
    else:
        out_value = 1 - vals[0] / np.linalg.norm([vals[1], vals[2]])
    return out_value



# def measure_planarity(st: Stream, pick: Pick, window_length):
#     st.rotate('->')
#     pass










from glob import glob
from importlib import reload
from useis.ai import model
from useis.ai import dataset

from tqdm import tqdm
import matplotlib.pyplot as plt

reload(model)
reload(dataset)

input_files = '/data_1/pick_dataset/*.pickle'

file_list = glob(input_files)

picker_dataset = dataset.PickingDataset(file_list)

train_dataset, test_dataset = picker_dataset.split(split_fraction=0.8)

picker = model.AIPicker()

epoch = 100
train_losses = []
test_losses = []
batch_size = 2000
for e in tqdm(range(0, epoch)):
    picker.train(train_dataset, batch_size=batch_size)
    predictions, test_loss_mean, test_loss_std = picker.validate(
        test_dataset, batch_size=batch_size)
    test_losses.append(test_loss_mean)
    predictions, train_loss_mean, train_loss_std = picker.validate(
        train_dataset, batch_size=batch_size)
    train_losses.append(train_loss_mean)

    print(f'test loss: {test_loss_mean:0.2f}\ntrain loss: {train_loss_mean:0.2f}')

picker.save('/data_1/ai_models/ai_picker.pickle')

plt.figure(1)
plt.clf()
plt.plot(train_losses, label='train losses')
plt.plot(test_losses, label='test losses')
plt.legend()
plt.show()
plt.pause(0.1)



sampling_rate = 5000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
image_width = 224
image_height = 224
buffer_image_fraction = 0.05


from useis.processors import classifier
from importlib import reload
from useis.ai import model
from uquake.core import read
from uquake.core.util.requests import download_file_from_url
from time import time
reload(classifier)

base_directory = '.'
project = 'test_project'
network = 'test_network'

event_classifier = classifier.Classifier2(base_directory, project, network)

reload(model)

if event_classifier.event_classifier is None:
    classifier_model = model.EventClassifier2.load()
    event_classifier.add_model(classifier_model)

test_seismic_event_url = \
    'https://www.dropbox.com/s/ld78tj7gtoebxuj/seismic_event_1.mseed?dl=1'
# test_blast_url =
seismic_event_mseed_io = download_file_from_url(test_seismic_event_url)
# blast_mseed_io = download_file_from_url()

st_event = read(seismic_event_mseed_io)
# st_blast = read(blast_mseed_io)

t0 = time()
print(event_classifier.predict(st_event))
t1 = time()

print(f'{t1 - t0}')

# event_classifier.add_model()


from uquake.grid.nlloc import (ModelLayer, LayeredVelocityModel,
                               VelocityGridEnsemble)
import numpy as np


def make_layered_model():

    origin = np.array([650200, 4766170, -500])
    dimensions = np.array([100, 101, 68])
    spacing = np.array([25, 25, 25])

    z = [1168, 459, -300, -500]
    vp_z = [4533, 5337, 5836, 5836]
    vs_z = [2306, 2885, 3524, 3524]

    p_layered_model = LayeredVelocityModel(project_code)
    s_layered_model = LayeredVelocityModel(project_code, phase='S')
    for (z_, vp, vs) in zip(z, vp_z, vs_z):
        layer = ModelLayer(z_, vp)
        p_layered_model.add_layer(layer)
        layer = ModelLayer(z_, vs)
        s_layered_model.add_layer(layer)

    vp_grid_3d = p_layered_model.to_3d_grid(network_code, dimensions, origin,
                                             spacing)
    vs_grid_3d = s_layered_model.to_3d_grid(network_code, dimensions, origin,
                                             spacing)
    return VelocityGridEnsemble(vp_grid_3d, vs_grid_3d)



from uquake import read, read_inventory
from useis.core.project_manager import ProjectManager
from uquake import __package_name__ as ns

inventory = read_inventory('OT_old.xml')

for i, station in enumerate(inventory[0]):
    old_station_code = station.code
    station.historical_code = station.historical_code.replace('-', '_')
    if 'UG' in station.historical_code:
        station_number = int(station.historical_code.split('_')[0][4:])
        location_code = int(station.historical_code.split('_')[1][0])
        inventory[0][i].code = f'UG{station_number:02d}'
        if 'ACC' in station.historical_code:
            # Accelerometer
            channel_prefix = 'GN'
            if station_number in [1, 2, 3, 4, 5]:
                location_code = 1
        else:
            # Geophone
            channel_prefix = 'GP'
            if station_number in [1, 2, 3, 4, 5]:
                location_code = 2
    else:
        station_number = int(station.historical_code.split('_')[0][2:])
        location_code = int(station.historical_code.split('_')[1][0])
        inventory[0][i].code = f'SS{station_number:02d}'

    for extra_key in inventory[0][i].extra.keys():
        inventory[0][i].extra[extra_key].namespace = ns

    for j, channel in enumerate(inventory[0][i]):
        channel_code = channel.code
        if ('UG' in inventory[0][i].code) and (location_code == 3):
            location_code = 1
        inventory[0][i][j].location_code = f'0{location_code}'
        inventory[0][i][j].code = f'{channel_prefix}{channel_code.upper()}'
        for extra_key in inventory[0][i][j].extra.keys():
            inventory[0][i][j].extra[extra_key].namespace = ns

inventory.write('OT.xml', format='STATIONXML')




from pathlib import Path
import pickle
from uquake.core.inventory import read_inventory
from uquake.core.logging import logger
from uquake.nlloc import (Srces)
from uquake.core.event import ConfidenceEllipsoid, OriginUncertainty
from uquake.grid import nlloc as nlloc_grid
from uquake.grid import read_grid
from time import time
import numpy as np
import os
import shutil
from ..settings.settings import Settings
from uquake.core.event import AttribDict
from uquake.grid.hdf5 import H5TTable
from .projection import Projection
from pydantic import conlist


def read_srces(fname):
    with open(fname, 'rb') as srces_file:
        return pickle.load(srces_file)


class ProjectManager(object):

    inventory_file_name = 'inventory.xml'

    def __init__(self, base_projects_path, project_name, network_code,
                 use_srces=False, **kwargs):
        """
        Interface to manage project and grids.

        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        :param use_srces: if True use the srces files instead of the the
        inventory file should both files be present (default=False)
        :Example:
        >>> from uquake.grid import nlloc as nlloc_grid
        >>> from uquake.nlloc import nlloc
        # initialize the project with the root path of the project,
        the project and the network names
        >>> project_name = 'test'
        >>> network_code = 'test'
        >>> root_path=''
        >>> pm = nlloc.ProjectManager(root_path, project_name, network_code)
        # this will initialize the project and create a specific directory
        # structure if the directories do not exist.

        # add an inventory file to the project
        >>> path_to_uquake_inventory_file = 'PATH/TO/THE/INVENTORY/FILE.xml'

        ..note :: the uquake inventory object inherit from the obspy inventory
                  and
                  behaves in a very similar way. It, however, differs from
                  its parent has it implements properties specific to local
                  earthquake who are either expressed using UTM coordinates
                  system or a local coordinates system that cannot necessarily
                  be translated to latitude and longitude. Beyond this minor
                  difference, uquake inventory structure also differs from
                  the standard inventory and implements a slightly different
                  hierarchy more representative of mine monitoring systems.
                  The uquake Inventory object hierarchy is as follows:
                 1. Inventory
                    1.1. Networks: A useis network. a network includes at
                                   least one data acquisition station.
                        1.1.1 Stations: A place where the data acquisition is
                                        performed. For instance, station
                                        usually includes power, communication
                                        and data acquisition equipment. One
                                        or more site can be connected to a
                                        station.
                            1.1.1.1 sites: A instrument converting a
                                             physical phenomenon to data either
                                             digital or analog. A site
                                             comprises one or more channel.
                                1.1.1.1.1 Channels: A channel is a the smallest
                                                    unit of measuring.


        >>> inventory = read_inventory(path_to_uquake_inventory_file)
        >>> pm.add_inventory(inventory)

        # alternatively, sites can be added to the using the srces object.
        # sites can be added this way from a nlloc.nlloc.Srces object using
        the .add_srces method.
        ..note:: srces stands for sources and it is the nomenclature used in
                 NonLinLoc. This might be a soure of confusion for the users.
                 In addition, what NonLinLoc refers to as station is called a
                 sites in this context.

        # srces object can be constructed from an inventory as follows:
        >>> srces = Srces.from_inventory(inventory)
        # alternatively, each sites can be added individually using the
        .add_site method. As follows:
        >>> srces = Srces()
        >>> x = 250
        >>> y = 250
        >>> z = 250
        >>> elevation = 0
        >>> srces.add_site('site label', x, y, z, elev=elevation)

        # Srces can be added to the project as follows:
        >>> pm.add_srces(srces)

        # add the velocity models to the project. P- and S-wave velocity models
        can be added separately from a nlloc.grid.VelocityGrid3D
        using the .add_velocity method or from a nlloc.grid.VelocityEnsemble
        object using the .add_velocities method

        >>> origin = [0, 0, 0]
        >>> spacing = [25, 25, 25]
        >>> dimensions = [100, 100, 100]

        >>> nlloc_grid.VelocityGrid3D()
        >>> vp = 5000  # P-wave velocity in m/s
        >>> vs = 3500  # S-wave velocity in m/s
        >>> p_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='P',
        >>>                                        value=5000)
        >>> s_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='S',
        >>>                                        value=5000 )
        >>> pm.add_velocity(p_velocity)
        >>> pm.add_velocity(s_velocity)
        # Alternatively
        >>> velocities = nlloc_grid.VelocityGridEnsemble(p_velocity,
        >>>                                              s_velocity)
        >>> pm.add_velocities(velocities)
        # Adding a velocity model of the velocity models triggers the
        # calculation of the travel time grids.
        # It is possible to manually trigger the calculation of the travel
        # time grid by invoking
        >>> pm.init_travel_time_grids()
        # this should not, however, be required.

        # prior to running the location, NonLinLoc need to be configured.
        # configuring NonLinLoc can be done using the nlloc.nlloc.NonLinLoc
        # object
        >>> nonlinloc = nlloc.NonLinLoc()
        # this will initialize the nonlinloc object sith default value. Those
        # value have been used to locate useis events in a volumes of
        # approximately 3000 m x 3000 m x 1500 m. The parameters should be
        # provide adequate results for volumes of similar scale but would need
        # to be adapted to smaller or larger volumes.

        """

        self.project_name = project_name
        self.network_code = network_code

        p_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(network_code,
                                                                  'P')
        s_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(network_code,
                                                                  'S')

        self.base_projects_path = Path(base_projects_path)
        base_project_path = Path(base_projects_path)
        root = base_project_path / project_name / network_code
        self.paths = {'base': base_project_path,
                      'root': base_project_path / project_name / network_code,
                      'archives': base_project_path / 'archives',
                      'inventory': root / 'inventory',
                      'config': root / 'config',
                      'velocities': root / 'velocities',
                      'times': root / 'times',
                      'hdf5_times': root / 'hdf5_times',
                      'ai_models': root / 'ai_models',
                      'debug': root / 'debug',
                      'index': root / 'index',
                      'databases' : root / 'databases'}

        self.paths = AttribDict(self.paths)

        self.files = {'inventory': self.paths.inventory / 'inventory.xml',
                      'srces': self.paths.inventory / 'srces.pickle',
                      'settings': self.paths.config / 'settings.toml',
                      'services_settings': self.paths.config /
                                           'services_settings.toml',
                      'projection_settings': self.paths.config /
                                             'projection_settings.toml',
                      'p_velocity': self.paths.velocities / p_vel_base_name,
                      's_velocity': self.paths.velocities / s_vel_base_name,
                      'hdf5_tt': self.paths.hdf5_times /
                                 'travel_times_tables.h5f',
                      'classification_model': self.paths.ai_models /
                                              'classification.pt',
                      'picker_model': self.paths.ai_models / 'picker.pickle'}

        self.files = AttribDict(self.files)
        self.databases = AttribDict()

        # create the directory if it does not exist
        self.paths.root.mkdir(parents=True, exist_ok=True)
        for key in self.paths.keys():
            path = self.paths[key]
            logger.info(path)
            path.mkdir(parents=True, exist_ok=True)

        # SETTINGS
        settings_path = Path(os.path.realpath(__file__)).parent.parent

        if not self.files.settings.is_file():
            settings_template = settings_path / \
                                    'settings/settings_template.toml'

            shutil.copyfile(settings_template, self.files.settings)

        if not self.files.projection_settings.is_file():
            settings_template = settings_path / \
                                'settings/projection_settings_template.toml'
            shutil.copyfile(settings_template, self.files.projection_settings)

        if not self.files.services_settings.is_file():
            settings_path = Path(os.path.realpath(__file__)).parent.parent
            settings_template = settings_path / \
                                'settings/services_settings_template.toml'

            shutil.copyfile(settings_template, self.files.services_settings)

        self.settings = Settings(str(self.paths.config))

        self.srces = None
        self.inventory = None

        if self.files.inventory.exists():
            self.inventory = read_inventory(str(self.files.inventory))
            self.srces = Srces.from_inventory(self.inventory)
        elif self.files.srces.exists():
            with open(self.files.srces, 'rb') as srces_file:
                self.srces = pickle.load(srces_file)

        self.p_velocity = None
        try:
            fname = str(self.paths.velocities / p_vel_base_name)
            self.p_velocity = read_grid(fname, format='NLLOC')
        except Exception as e:
            logger.warning(e)

        self.s_velocity = None
        try:
            fname = str(self.paths.velocities / s_vel_base_name)
            self.s_velocity = read_grid(fname, format='NLLOC')
        except Exception as e:
            logger.warning(e)

        self.velocities = None
        if (self.p_velocity is not None) and (self.s_velocity is not None):
            self.velocities = nlloc_grid.VelocityGridEnsemble(self.p_velocity,
                                                              self.s_velocity)

        self.paths.times = self.paths.root / 'times'
        self.paths.times.mkdir(parents=True, exist_ok=True)
        file_list = list(self.paths.times.glob('*time*'))

        self.travel_times = None
        try:
            self.travel_times = nlloc_grid.TravelTimeEnsemble.from_files(
                self.paths.times)
        except Exception as e:
            logger.error(e)

    @property
    def htt(self):
        return self.hdf5_tt

    @property
    def hdf5_tt(self):
        return self.travel_times.to_hdf5(self.files.hdf5_tt)

    @property
    def projection(self):
        offset = self.settings.projection.local.offset.to_list()
        return Projection(global_epsg_code=
                          self.settings.projection.GLOBAL.epsg,
                          local_epsg_code=self.settings.projection.local.epsg,
                          offset=offset)

    def set_global_epsg_code(self, epsg_code: int):
        self.settings.projection.GLOBAL.epsg = epsg_code
        self.projection.write(self.files.projection_settings)

    def set_local_epsg_code(self, epsg_code: int):
        self.settings.projection.local.epsg = epsg_code
        self.projection.write(self.files.projection_settings)

    def set_projection_offset(self, offset: conlist(float, min_items=3,
                                                    max_items=3)):
        self.settings.projection.local.offset = offset
        self.projection.write(self.files.projection_settings)

    @property
    def global_epsg_code(self):
        return self.projection.global_epsg_code

    @global_epsg_code.setter
    def global_epsg_code(self, value):
        self.projection.global_epsg_code = value
        self.projection.write(self.files.projection_settings)

    @property
    def local_epsg_code(self):
        return self.projection.local_epsg_code

    @local_epsg_code.setter
    def local_epsg_code(self, value):
        self.projection.local_epsg_code = value
        self.projection.write(self.files.projection_settings)

    @property
    def projection_offset(self):
        return self.projection.offset

    @local_epsg_code.setter
    def local_epsg_code(self, value: conlist(float, min_items=3, max_items=3)):
        self.projection.offset = value

        self.projection.write(self.files.projection_settings)

    def transform_to_global(self, easting: float = 0, northing: float = 0,
                            z: float = 0):
        return self.projection.transform_to_global(easting=easting,
                                                   northing=northing,
                                                   z=z)

    def transform_to_local(self, longitude: float = 0, latitude: float = 0,
                           z: float = 0):
        return self.projection.transform_to_local(longitude=longitude,
                                                  latitude=latitude)

    @staticmethod
    def exists(project_path, project_name, network_code):
        project_path = Path(project_path) / project_name / network_code
        return project_path.exists()

    def init_travel_time_grids(self, multi_threaded=True):
        """
        initialize the travel time grids
        """

        has_velocity = self.has_p_velocity() or self.has_s_velocity()

        abort = not (has_velocity and self.has_inventory())

        if not self.has_p_velocity():
            logger.warning('No P-wave velocity found in project')

        if not self.has_s_velocity():
            logger.warning('No S-wave velocity found in project')

        if not self.has_inventory():
            logger.warning('No inventory file found in project')

        if abort:
            logger.warning('travel time grids will not be calculated')
            return

        logger.info('initializing the travel time grids')
        t0 = time()
        seeds = self.srces.locs
        seed_labels = self.srces.labels

        tt = self.velocities.to_time(seeds, seed_labels,
                                     multi_threaded=multi_threaded)

        self.travel_times = tt

        for fle in self.paths.times.glob('*time*'):
            fle.unlink(missing_ok=True)

        self.travel_times.write(self.paths.times)
        self.travel_times.write_hdf5(self.files.hdf5_tt)
        t1 = time()
        logger.info(f'done initializing the travel time grids in '
                    f'{t1 - t0:0.2f} seconds')

    def add_inventory(self, inventory, create_srces_file: bool = True,
                      initialize_travel_time: bool = True):
        """
        adding a inventory object to the project
        :param inventory: station xml inventory object
        :type inventory: uquake.core.inventory.Inventory
        :param create_srces_file: if True create or replace the srces file
        (Default: True)
        :param initialize_travel_time: if True the travel time grids are
        initialized (default: True)
        :return:
        """

        inventory.write(str(self.files.inventory))
        self.srces = Srces.from_inventory(inventory)
        self.inventory = inventory
        if create_srces_file:
            with open(self.files.srces, 'wb') as srces_file:
                pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grids()
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_inventory_from_file(self, inventory_file: str, format='StationXML',
                                create_srces_file: bool = True,
                                initialize_travel_time: bool = True,
                                xy_from_lat_lon: bool = False,
                                input_projection=4326,
                                output_projection=None):
        """
        adding a inventory object to the project from a file
        :param inventory_file: station xml inventory object
        :type inventory_file: uquake.core.inventory.Inventory
        :param format: file format for the inventory file
        :param create_srces_file: if True create or replace the srces file
        (Default: True)
        :param initialize_travel_time: if True the travel time grids are
        initialized (default: True)
        :param xy_from_lat_lon: convert latitude and longitude in the file to
        cartesian coordinates
        :param input_projection: input projection (default=4326 for typical
        latitude and longitude)
        :param output_projection: output projection (default None)
        :return:
        """

        if output_projection is not None:
            inventory = read_inventory(inventory_file, format=format,
                                       xy_from_lat_lon=xy_from_lat_lon,
                                       input_projection=input_projection,
                                       output_projection=output_projection)
        else:
            inventory = read_inventory(inventory_file, format=format,
                                       xy_from_lat_lon=xy_from_lat_lon,
                                       input_projection=input_projection)

        self.add_inventory(inventory, create_srces_file=create_srces_file,
                           initialize_travel_time=initialize_travel_time)

    def add_srces(self, srces, force=False, initialize_travel_time: bool=True,
                  multi_threaded: bool=True):
        """
        add a list of sources to the projects
        :param srces: list of sources or sites
        :param force: force the insertion of the srces object if an inventory
        file is present
        :param initialize_travel_time: if True, initialize the travel time
        grid
        :type srces: Srces
        :param multi_threaded: if true, use multi-threading
        :type multi_threaded: bool

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        if not isinstance(srces, Srces):
            raise TypeError(f'Expecting type {Srces}, given '
                            f'{type(srces)}')

        if self.inventory is not None:
            logger.warning('The project already has an inventory file!')
            if not force:
                logger.warning('exiting...')
                return
            else:
                logger.warning('the force flag value is True, srces object '
                               'will be added.')

        self.srces = srces
        with open(self.files.srces, 'wb') as srces_file:
            pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grids(multi_threaded=multi_threaded)
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_velocities(self, velocities, initialize_travel_times=False):
        """
        add P- and S-wave velocity models to the project
        :param velocities: velocity models
        :type velocities: nlloc.grid.VelocityEnsemble
        :param initialize_travel_times: if True, initialize the travel time
        grid

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        # velocities.write(path=self.paths.velocities)

        self.velocities = velocities

        for key in velocities.keys():
            self.add_velocity(velocities[key])

        if initialize_travel_times:
            self.init_travel_time_grids()

    def velocity_versioning_handler(self, phase):
        """
        returns the version number of the to be archived velocity model
        :return: int
        """
        file_list = [fle for fle in self.paths.archives.glob(f'*.{phase}.*')]
        # file_list = os.listdir(self.paths.archives / )
        if not file_list:
            return 1

        exts = []
        for element in file_list:
            ext = int(element.split('_')[-1])
            exts.append(ext)

        return np.max(exts) + 1

    def add_velocity(self, velocity, initialize_travel_times=False):
        """
        add P- or S-wave velocity model to the project
        :param velocity: p-wave velocity model
        :type velocity: uquake.grid.nlloc.VelocityGrid3D
        :param initialize_travel_times: if true initialize the travel time grids

        ..warning:: travel time should be initialized when the sites/srces
        are updated. Not doing so, may cause the sites/source and the
        travel time grids to be incompatible.
        """

        if velocity.phase.upper() == 'P':
            self.p_velocity = velocity
        elif velocity.phase.upper() == 'S':
            self.s_velocity = velocity

        velocity.write(self.paths.velocities)

        if initialize_travel_times:
            try:
                self.init_travel_time_grids()
            except Exception as e:
                logger.warning(e)

    def list_active_projects(self):
        """
        List all active projects
        :return: list of active projects
        """
        projects = []
        for project in self.base_projects_path.glob('*'):
            if project.name == 'archives':
                continue
            if project.is_dir():
                projects.append(project.name)

        return projects

    def list_active_networks_project(self, project=None):
        """
        List all active networks
        :param project: the project for which to list the networks
        :return: a list of networks
        """
        if project is None:
            project = self.project_name

        project_directory = self.base_projects_path / project

        if not project_directory.exists():
            logger.warning(f'the project {project} does not exists')
            return

        networks = []
        for network in project_directory.glob('*'):
            if not network.is_dir():
                continue
            networks.append(network.name)

        return networks

    def list_all_active_networks(self):
        """
        list all active projects and networks
        :return: a dictionary of list
        """

        out_dict = {}
        for project in self.list_active_projects():
            networks = self.list_active_networks_project(project=project)
            out_dict[project] = networks

        return out_dict

    def list_all_archived_networks(self):
        """
        List all archived projects and networks
        :return: dictionary of list where the keys are the projects name
        """

        out_dict = {}
        for project in self.paths.archives.glob('*'):
            if not project.is_dir():
                continue
            archived_project_dir = self.paths.archives / project.name
            out_dict[project.name] = []
            for network in archived_project_dir.glob('*'):
                if not network.is_dir():
                    continue
                out_dict[project.name].append(network.name)

        return out_dict

    def has_p_velocity(self):
        if self.p_velocity:
            return True
        return False

    def has_s_velocity(self):
        if self.s_velocity:
            return True
        return False

    def has_inventory(self):
        if self.inventory or self.srces:
            return True
        return False

    def clean_stream(self, stream):
        if 'black_list'.upper() in self.settings.keys():
            for sensor in self.settings.black_list:
                station = sensor.split('.')[0]
                location = sensor.split('.')[1]
                for tr in stream.select(station=station, location=location):
                    stream.remove(tr)

        return stream





from ..core.project_manager import ProjectManager
import pandas as pd

from pathlib import Path
import os

from uquake.core import read, UTCDateTime
from multiprocessing import Pool, cpu_count
from uquake.core.logging import logger
import itertools
from tqdm import tqdm
import numpy as np
import pandas as pd
from uquake.core.stream import Stream


def extract_trace_info_mseed(mseed_file):
    try:
        st = read(mseed_file)
    except Exception as e:
        logger.warning(e)
        return []
    output_list = []
    for tr in st:

        output_dict = {'mseed_file': mseed_file,
                       'start_time': tr.stats.starttime.datetime,
                       'start_time_timestamp': tr.stats.starttime.timestamp,
                       'end_time': tr.stats.endtime.datetime,
                       'end_time_timestamp': tr.stats.endtime.timestamp,
                       'network': tr.stats.network,
                       'station': tr.stats.station,
                       'location': tr.stats.location,
                       'channel': tr.stats.channel}

        output_list.append(output_dict)

    return output_list


def change_channel_name(mseed_file):
    from uquake.core.stream import Stream
    try:
        st = read(mseed_file)
    except Exception as e:
        logger.warning(e)
        return []
    trs = []
    for tr in st.copy():
        if 'N' == tr.stats.channel[-1]:
            tr.stats.channel = tr.stats.channel.replace('N', '1')
        elif 'E' == tr.stats.channel[-1]:
            tr.stats.channel = tr.stats.channel.replace('E', '2')
        else:
            tr.stats.channel = tr.stats.channel.replace('Z', '3')

        trs.append(tr)

    output_mseed = Path(str(mseed_file).replace('data_sept',
                                                'data_sept_reprocessed'))

    output_mseed.parent.mkdir(exist_ok=True, parents=True)

    Stream(traces=trs).write(output_mseed)
    return mseed_file


class ContinuousWaveformDataExtractor(ProjectManager):
    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str, data_path: Path,
                 create_waveform_file_index: bool = False,
                 multiprocessing_indexing: bool = True,
                 cpu_utilisation: float = 0.9):

        super().__init__(base_projects_path, project_name, network_code)
        self.paths.waveform_data = data_path

        self.files.waveform_file_index = self.paths.index / \
                                'continuous_waveform_file.index'

        self.waveform_file_index = None
        self.__cpu_count__ = cpu_count()
        self.__cpu_utilisation__ = cpu_utilisation
        if self.files.waveform_file_index.exists():
            logger.info(f'reading file index '
                        f'{self.files.waveform_file_index}')
            self.waveform_file_index = pd.read_csv(
                self.files.waveform_file_index,
                parse_dates=['start_time', 'end_time'])

        if self.waveform_file_index is None:
            if create_waveform_file_index:
                self.create_index()
            else:
                logger.warning('waveform file index does not exist and will '
                                'not be created')

    def create_index(self, multiprocessing: bool = True):

        num_threads = int(np.ceil(self.__cpu_utilisation__ * cpu_count()))
        #
        # file_index = {'starttime': [],
        #               'endtime': [],
        #               'network': [],
        #               'station': [],
        #               'location': [],
        #               'channel': []}

        wf_files = []
        for root, path, waveform_files in os.walk(self.paths.waveform_data):
            root = Path(root)
            for waveform_file in waveform_files:
                wf_files.append(root / waveform_file)

        if multiprocessing:
            with Pool(num_threads) as p:
                r = list(tqdm(p.imap(extract_trace_info_mseed,
                                      wf_files), total=len(wf_files)))
            rtmp = list(itertools.chain(*r))
            results = {key: [dic[key] for dic in rtmp]
                       for key in rtmp[0].keys()}

        logger.info('writing the file index')
        self.waveform_file_index = pd.DataFrame(results)
        self.waveform_file_index['start_time'] = \
            pd.to_datetime(self.waveform_file_index['start_time'])

        self.waveform_file_index['end_time'] = \
            pd.to_datetime(self.waveform_file_index['end_time'])

        self.waveform_file_index.to_csv(self.files.waveform_file_index)

    def get_waveforms(self, start_time, duration, network=None, station=None,
                     location=None, channel=None) -> Stream:

        start_time = UTCDateTime(start_time)

        end_time = start_time + duration
        it = self.waveform_file_index[
            (self.waveform_file_index['start_time_timestamp']
             <= end_time.timestamp)]
        it2 = it[it['end_time_timestamp'] > start_time.timestamp]

        mseed_files = np.unique(it2['mseed_file'])

        if len(mseed_files) == 0:
            raise Exception('the request did not match any file in the index')

        trs = []
        for mseed_file in mseed_files:
            st = read(mseed_file)
            for tr in st:
                trs.append(tr)

        st_out = Stream(traces=trs).merge()

        st2 = st_out.select(network=network, station=station,
                            location=location, channel=channel)
        return st2.trim(starttime=start_time, endtime=end_time)








from ..core.project_manager import ProjectManager
from uquake.core.stream import Stream
from uquake.core.event import Event
import numpy as np
from uquake.waveform.simple_mag import moment_magnitude
from pathlib import Path
import os
import shutil


class Magnitude(ProjectManager):
    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str):
        """
        Based object for the magnitude calculation
        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        """

        super().__init__(base_projects_path, project_name, network_code)

        self.files.magnitude_settings = self.paths.config / \
                                        'magnitude_settings.toml'

        if not self.files.magnitude_settings.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                                '../settings/magnitude_settings_template.toml'

            shutil.copyfile(settings_template,
                            self.files.magnitude_settings)

            super().__init__(base_projects_path, project_name, network_code)


class MwPE(Magnitude):
    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str):
        """
        Based object for the magnitude calculation
        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        """

        super().__init__(base_projects_path, project_name, network_code)

    def predict(self, st: Stream, event_location: np.array,
                use_distance_along_rays=True):
        """

        :param st: waveform
        :type st: :py:class:uquake.core.stream.Stream:
        :param event_location: event location
        :type event_location: :py:class:numpy.array:
        :param use_distance_along_rays: Use the distance along the rays if True
        use the straight line distance if False. Default True.
        :type use_distance_along_rays: bool
        :return: the magnitude value estimate
        :rtype: float
        """

        a = self.settings.MwPE.a
        b = self.settings.MwPE.c

        min_frequency = 0
        for tr in st:
            trace_time = (tr.stats.endtime - tr.stats.starttime) * 0.9
            if 1 / trace_time > min_frequency:
                min_frequency = 1 / trace_time

        min_frequency_1 = min_frequency
        min_frequency_2 = 10 ** (np.log10(min_frequency) + 0.2)

        pre_filt = [min_frequency_1, min_frequency_2]

        max_frequency_2 = self.settings.MwPE.max_frequency
        max_frequency_1 = 10 ** (np.log10(max_frequency_2) - 0.2)

        pre_filt = [min_frequency_1, min_frequency_2,
                    max_frequency_1, max_frequency_2]

        water_level = self.settings.MwPE.water_level

        st.attach_response(self.inventory)
        st.remove_response(inventory=self.inventory, pre_filt=pre_filt,
                           water_level=water_level, output='VEL')

        if use_distance_along_rays:
            rays = self.travel_times.ray_tracer(event_location,
                                                seed_labels='current_event')

        magnitudes = []
        for network in self.inventory:
            for station in network:
                st2 = st.copy().select(network=network, station=station)
                if len(st2) == 1:
                    amplitude = np.max(st2[0].data) * np.sqrt(3)
                else:
                    amplitude = np.max(st2.composite()[0].data)

        networks = self.inventory


class MomentMagnitude(Magnitude):
    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str):
        """
        Based object for the magnitude calculation
        :param base_projects_path: base project path
        :type base_projects_path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        """

        super().__init__(base_projects_path, project_name, network_code)

    def predict(self, st: Stream, event: Event, only_triaxial=True,
                density=2700, min_dist=20, win_length=0.04, len_spectrum=2**12,
                clipped_fraction=0.1, max_frequency=600,
                preferred_origin_only=True):
        """
        Calculate the moment magnitude for a given event
        :param st: waveform
        :type st: :py:class:uquake.core.stream.Stream:
        :param event: event location
        :type event: :py:class:numpy.array:
        :return: the magnitude value estimate
        :rtype: float
        :param only_triaxial: whether only triaxial sensor are used in the
        magnitude calculation (optional) (not yet implemented)
        :type only_triaxial: bool
        :param density: density in kg / m**3 (assuming homogeneous for now)
        :type density: float
        :param win_length: length of the window in second in which magnitude is
        calculated
        :type win_length: float
        :param min_dist: minimum distance between sensor an event to allow
        magnitude calculation
        :param len_spectrum: length of the spectrum
        :param clipped_fraction: allowed clipped fraction (fraction of the
        signal equal to the min or the max.
        :param max_frequency: maximum frequency used in the calculation on
        magnitude. After a certain frequency, the noise starts to dominate the
        signal and the biases the calculation of the magnitude and corner
        frequency.
        :param preferred_origin_only: calculates the magnitude for the
        preferred_origin only
        :rtype: uquake.core.event.Catalog
        """

        mag = moment_magnitude(st, event, inventory, self.p_velocity,
                               self.s_velocity, only_triaxial=only_triaxial,
                               density=density, min_dist=min_dist,
                               win_length=win_length,
                               len_spectrum=len_spectrum,
                               clipped_fraction=clipped_fraction,
                               max_frequency=max_frequency,
                               preferred_origin_only=preferred_origin_only)

        return mag





            





from sqlalchemy import (create_engine, Column, Integer, String, Float, Boolean, MetaData,
                        distinct, Table, select)
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
import sqlite3
from ipdb import set_trace
import pandas as pd
import sqlite3

# Define a declarative base for the ORM
Base = declarative_base()


# Define a class to represent the record table
class Record(Base):
    __tablename__ = 'records'
    id = Column(Integer, primary_key=True)
    event_id = Column(String)
    spectrogram_filename = Column(String)
    station = Column(String)
    location = Column(String)
    channel = Column(String)
    channel_id = Column(Integer)
    magnitude = Column(Float)
    duration = Column(Float)
    end_time = Column(String)
    sampling_rate = Column(Float)
    categories = Column(String)
    original_event_type = Column(String)
    mseed_file = Column(String)
    sensor_type = Column(String)
    bounding_box_start = Column(Integer)
    bounding_box_end = Column(Integer)
    synthetic = Column(Boolean)

    def __repr__(self):
        str = ''
        for key in self.__dict__.keys():
            str += f'{key}: {self.__dict__[key]}\n'

        return str


class DBManager(object):
    def __init__(self, db_url):
        self.db_url = db_url
        self.db_path = db_url[10:]

        self.engine = create_engine(self.db_url)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()
        self.metadata = MetaData()
        Base.metadata.create_all(self.engine)
        self.table_name = Record.__tablename__

    def add_record(self, event_id=None, spectrogram_filename=None,
                   station=None, location=None, channel=None, channel_id=None,
                   magnitude=None, duration=None, end_time=None,
                   sampling_rate=None, categories=None, original_event_type=None,
                   mseed_file=None, sensor_type=None, bounding_box=None,
                   synthetic=False):

        if bounding_box is not None:
            bounding_box_start = bounding_box[0]
            bounding_box_end = bounding_box[1]
        else:
            bounding_box_start = 0
            bounding_box_end = 0

        record = Record(event_id=event_id,
                        spectrogram_filename=spectrogram_filename,
                        station=station,
                        location=location,
                        channel=channel,
                        channel_id=channel_id,
                        magnitude=magnitude,
                        duration=duration,
                        end_time=end_time,
                        sampling_rate=sampling_rate,
                        categories=categories,
                        original_event_type=original_event_type,
                        mseed_file=mseed_file,
                        sensor_type=sensor_type,
                        bounding_box_start=bounding_box_start,
                        bounding_box_end=bounding_box_end,
                        synthetic=synthetic)
        session = self.Session()

        session.add_all([record])
        session.commit()
        session.close()

    # def clear_database(self):
    #     self.metadata.reflect()
    #
    #     for table in reversed(metadata.sorted_tables):
    #         conn = engine.connect()
    #         conn.execute(table.delete())
    #         conn.close()

    def clear_database(self):
        # Get the table object based on the table name stored in the class

        table = Table(self.table_name, self.metadata, sqlite3_autoload=True,
                      autoload_with=self.engine)

        # Delete all records from the table
        delete_statement = table.delete()
        self.session.execute(delete_statement)
        self.session.commit()

    def event_exists(self, event_id):
        session = self.Session()
        query = session.query(Record).filter_by(event_id=event_id)
        record_exists = session.query(query.exists()).scalar()
        session.close()
        return record_exists

    def filter(self, **kwargs):
        session = self.Session()
        query = session.query(Record).filter_by(**kwargs)
        session.close()
        return query

    def all_unique(self, field):
        records = Record.__table__
        query = select(getattr(records.c, field))
        session = self.Session()
        results = session.execute(query).fetchall()
        unique_records = list(set(result[0] for result in results))
        session.close()
        return unique_records

    def to_pandas(self):
        con = sqlite3.connect(self.db_path)
        query = f'SELECT * FROM {self.table_name}'
        df = pd.read_sql_query(query, con)
        con.close()
        return df
        # return pd.read_sql_table(Record.__tablename__, self.engine)

    @property
    def categories(self):
        session = self.Session()
        categories = [category[0] for category in
                      session.query(distinct(Record.categories)).all()]
        session.close()
        return categories







import pandas
from uquake.core import read, read_events, read_inventory, UTCDateTime
from uquake.core.trace import Trace
from uquake.core.logging import logger
from pathlib import Path
import pandas as pd
from multiprocessing import Pool, cpu_count
import numpy as np
from tqdm import tqdm
import pickle
from useis.core.project_manager import ProjectManager
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
import matplotlib.pyplot as plt
import torch
from uquake.waveform.pick import calculate_snr
from PIL import Image, ImageOps
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from dataset import spectrogram


def get_waveform(waveform_file, inventory):
    st = read(waveform_file)
    for i, tr in enumerate(st):
        for sensor in inventory.sensors:
            if sensor.alternate_code == tr.stats.station:
                st[i].stats.network = inventory[0].code
                st[i].stats.station = sensor.station.code
                st[i].stats.location = sensor.location_code
                for channel in sensor.channels:
                    if tr.stats.channel in channel.code:
                        st[i].stats.channel = channel.code
                        break
                break
    return st


def get_event_id(filename):
    cat = read_events(filename)
    return str(filename.stem), cat[0].resource_id.id


event_type_lookup = {'anthropogenic event': 'noise',
                     'acoustic noise': 'noise',
                     'reservoir loading': 'noise',
                     'road cut': 'noise',
                     'controlled explosion': 'blast',
                     'quarry blast': 'blast',
                     'earthquake': 'seismic event',
                     'sonic boom': 'noise',
                     'collapse': 'noise',
                     'other event': 'noise',
                     'thunder': 'noise',
                     'induced or triggered event': 'noise',
                     'explosion': 'blast',
                     'experimental explosion': 'blast'}

input_data_dir = Path('/data_1/ot-reprocessed-data/')
output_data_dir = Path('/data_1/classification_dataset_2')
output_data_dir.mkdir(parents=True, exist_ok=True)
sampling_rate = 6000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
image_width = 128
image_height = 128
buffer_image_fraction = 0.05

buffer_image_sample = int(image_width * buffer_image_fraction)

hop_length = int(sequence_length_second * sampling_rate //
                 (image_width + 2 * buffer_image_sample))


# def spectrogram(trace: Trace, perturbation_sample: float = None):
#
#     mel_spec = MelSpectrogram(sample_rate=sampling_rate,
#                               n_mels=image_height,
#                               hop_length=hop_length,
#                               power=1,
#                               pad_mode='reflect',
#                               normalized=True)
#
#     amplitude_to_db = AmplitudeToDB()
#
#     trace = trace.taper(max_length=0.01, max_percentage=0.05)
#     # plt.figure(3)
#     # plt.clf()
#     # plt.plot(trace.data)
#
#     n_pts = sequence_length_second * sampling_rate
#
#     target_sequence_length = int(
#         2 ** (np.ceil(np.log(n_pts) / np.log(2))))
#     n_zeros = target_sequence_length - len(trace.data)
#
#     data = np.hstack((trace.data, np.zeros(n_zeros)))
#
#     data = np.roll(data, perturbation_sample)
#
#     # data = data[:sequence_length_second * sampling_rate]
#
#     torch_data = torch.tensor(data).type(torch.float32)
#
#     spec = (mel_spec(torch_data) + 1e-3)[:, buffer_image_sample:
#                                          buffer_image_sample
#                                          + image_width]
#     spec_db = amplitude_to_db(spec.abs())
#     spec_db = spec_db - spec_db.min()
#     spec_db = (spec_db / spec_db.max() * 255).type(torch.uint8)
#     img = Image.fromarray(np.array(spec_db.tolist()).astype(
#         np.uint8))
#
#     return img


def create_training_image(waveform_file):
    # logger.info('here')
    waveform_file_path = input_data_dir / waveform_file
    cat = read_events(waveform_file_path.with_suffix('.xml'))
    st = read(waveform_file_path)

    if cat[0].preferred_origin() is not None:
        origin = cat[0].preferred_origin()
    else:
        origin = cat[0].origins[-1]

    if origin.evaluation_mode == 'automatic':
        return

    event_time = origin.time.timestamp

    try:
        st = st.detrend('demean').detrend(
            'linear').taper(max_percentage=0.1,
                            max_length=0.01).resample(
            sampling_rate=sampling_rate)
    except Exception as e:
        logger.error(e)
        trs = []
        for tr in st:
            if np.nan in tr.data:
                continue
            trs.append(tr.copy())
        st.traces = trs

    trs = []
    logger.info(event_type_lookup[cat[0].event_type])
    if event_type_lookup[cat[0].event_type] == 'seismic event':
        for arrival in cat[0].preferred_origin().arrivals:
            site = arrival.pick.site
            for tr in st.select(site=site).copy():
                # filtering out low frequency before calculating the SNR
                tr = tr.filter('highpass', freq=100)
                snr = calculate_snr(tr, arrival.pick.time,
                                    pre_wl=20e-3,
                                    post_wl=20e-3)
                if arrival.pick.evaluation_mode == 'automatic':
                    if snr < snr_threshold:
                        continue
                tr.trim(starttime=tr.stats.starttime,
                        endtime=tr.stats.starttime +
                        sequence_length_second,
                        pad=True,
                        fill_value=0)
                trs.append(tr.copy().resample(sampling_rate=
                                              int(sampling_rate)))

    elif event_type_lookup[cat[0].event_type] == 'blast':
        for tr in st:
            if (np.max(np.abs(tr.data)) / np.std(tr.data)) < 6:
                continue
            trs.append(tr.trim(endtime=tr.stats.starttime +
                               sequence_length_second))

    else:
        for tr in st:
            trs.append(tr.trim(endtime=tr.stats.starttime +
                               sequence_length_second))

    ct = 0
    for tr in trs:
        try:
            spec = spectrogram(tr)
        except Exception as e:
            logger.error(e)
            ct += 1
            logger.info(f'{ct} exceptions')
            return
        spec = (spec / np.max(spec)) * 255
        img = Image.fromarray(np.array(spec.tolist()).astype(np.uint8))

        filename = f'{event_time:0.0f}_{tr.stats.site}_' \
                   f'{tr.stats.channel}.jpg'

        out_dir = (output_data_dir /
                   event_type_lookup[cat[0].event_type])

        out_dir.mkdir(parents=True, exist_ok=True)

        out_file_name = out_dir / filename
        ImageOps.grayscale(img).save(out_file_name, format='JPEG')


if __name__ == '__main__':
    file_list = [f for f in input_data_dir.glob('*.mseed')]
    for f in tqdm(file_list):
        create_training_image(f)

    # file_list = [f for f in input_data_dir.glob('*.mseed')]
    # with Pool(4) as pool:
    #     event_list = list(tqdm(pool.map(create_training_image, file_list),
    #                            total=len(file_list)))
    #     # with tqdm(total=len(file_list)) as p_bar:
    #     #     for i, _ in enumerate(pool.imap(create_training_image, file_list)):
    #     #         logger.info('here')
    #     #         p_bar.update()



from pydantic import BaseModel
from typing import Optional, List
from enum import Enum
from datetime import datetime
import uquake
from uquake.core.event import Pick as UQPick
from .base import Coordinates3D, Angle
from uuid import UUID


# class Phase(BaseModel, Enum):
#     P = "P"
#     S = "S"


class WaveformStreamID(BaseModel):
    station_code: Optional[str]
    channel_code: Optional[str]
    location_code: Optional[str]


class QuantityError(BaseModel):
    uncertainty: Optional[float]
    lower_uncertainty: Optional[float]
    upper_uncertainty: Optional[float]
    confidence_level: Optional[float]


class PickOnset(str, Enum):
    emergent = "emergent"
    impulsive = "impulsive"
    questionable = "questionable"


class Phase(str, Enum):
    P = 'P'
    S = 'S'


class PickPolarity(str, Enum):
    positive = "positive"
    negative = "negative"
    undecidable = "undecidable"


class EvaluationMode(str, Enum):
    manual = "manual"
    automatic = "automatic"


class EvaluationStatus(str, Enum):
    preliminary = "preliminary"
    confirmed = "confirmed"
    reviewed = "reviewed"
    final = "final"
    rejected = "rejected"
    reported = "reported"


class CreationInfo(BaseModel):
    agency_id: Optional[str]
    agency_uri: Optional[str]
    author: Optional[str]
    author_uri: Optional[str]
    creation_time: Optional[datetime]
    version: Optional[str]


class ResourceIdentifier(BaseModel):
    fixed: bool = False
    _prefix: str = 'smi:local'
    _uuid: str

    class Config:
        orm_mode = True


class SimpleArrival(BaseModel):
    time: datetime
    station_code: str
    location_code: str
    channel_code: str
    phase: Phase
    onset: PickOnset
    evaluation_mode: EvaluationMode
    evaluation_status: EvaluationStatus

    class Config:
        orm_mode = True

    @classmethod
    def from_uquake_pick(cls, pick: uquake.core.event.Pick):
        """
        convert a uquake Pick object into a SimpleArrival
        :param pick: a pick object
        :type pick: uquake.core.event.Pick
        :return:  a SimpleArrival
        :rtype: SimpleArrival
        """
        return cls(time=pick.time, station_code=pick.waveform_id.station_code,
                   location_code=pick.waveform_id.location_code,
                   channel_code=pick.waveform_id.channel_code,
                   phase=pick.phase_hint, onset=pick.onset,
                   evaluation_mode=pick.evaluation_mode,
                   evaluation_status=pick.evaluation_status)

    def to_uquake_pick(self) -> uquake.core.event.Pick:
        """
        convert to uquake Pick
        :return: uquake pick
        :rtype: uquake.core.event.Pick
        """
        waveform_id = WaveformStreamID(station_code=self.station_code,
                                       channel_code=self.channel_code,
                                       location_code=self.location_code)
        return UQPick(time=self.time, waveform_id=waveform_id,
                      phase_hint=self.phase, onset=self.onset,
                      evaluation_mode=self.evaluation_mode,
                      evaluation_status=self.evaluation_status)


class Ray(BaseModel):
    nodes: List[Coordinates3D]
    site_code: Optional[str]
    arrival_id: Optional[str]
    phase: Optional[Phase]
    azimuth: Optional[Angle]
    takeoff_angle: Optional[Angle]
    travel_time: Optional[float]
    earth_model_id: Optional[str]

    class config:
        from_orm = True

    @classmethod
    def from_uquake(cls, uq_ray: uquake.core.event.Ray):
        nodes = []
        for node in uq_ray.nodes:
            nodes.append(Coordinates3D(node[0], node[1], node[2]))
        return cls(nodes=nodes, site_code=uq_ray.site_code,
                   arrival_id=uq_ray.arrival_id, phase=uq_ray.phase,
                   azimuth=uq_ray.azimuth, takeoff_angle=uq_ray.takeoff_angle,
                   travel_time=uq_ray.travel_time,
                   earth_model_id=uq_ray.earth_model_id)



from uquake.nlloc.nlloc import (Control, GeographicTransformation,
                                LocSearchOctTree, LocationMethod,
                                GaussianModelErrors, LocQual2Err)

nlloc_control = {'control': Control(message_flag=1),
                 'transformation': GeographicTransformation(),
                 'locsig': 'Organisation Name',
                 'loccom': '',
                 'locsearch': LocSearchOctTree(20, 20, 20, 1e-6, 50000,
                                               1000, 0, 1),
                 'locmeth': LocationMethod('EDT_OT_WT', 9999.0, 4,
                                           -1, -1, -1, 0, 0),
                 'locgau': GaussianModelErrors(1e-2, 1e-2),
                 'locqual2err': LocQual2Err(1e-4, 1e-4, 1e-4, 1e-4,
                                            1e-4)
                 }



event_type_lookup = {'anthropogenic event': 'noise',
                     'acoustic noise': 'noise',
                     'reservoir loading': 'noise',
                     'road cut': 'noise',
                     'controlled explosion': 'blast',
                     'quarry blast': 'blast',
                     'earthquake': 'seismic event',
                     'sonic boom': 'noise',
                     'collapse': 'noise',
                     'other event': 'noise',
                     'thunder': 'impulsive noise',
                     'induced or triggered event': 'noise',
                     'explosion': 'blast',
                     'experimental explosion': 'blast'}


import pandas as pd

from ..core.project_manager import ProjectManager
from uquake.grid.nlloc import VelocityGrid3D, VelocityGridEnsemble
from uquake.nlloc.nlloc import Srces, Site, Observations
from uquake.core.event import (Catalog, Origin, Arrival, Pick,
                               WaveformStreamID, Ray, ResourceIdentifier)
from uquake.core import UTCDateTime
from uquake.core.logging import logger
import numpy as np
from multiprocessing import Pool, cpu_count
from tqdm import tqdm
import dataclasses
from pydantic import BaseModel, conlist
from enum import Enum
from pydantic.typing import List, Union
from datetime import datetime
from uuid import uuid4
from functools import partial
import scipy as sc
from scipy.sparse import csr_matrix
from ..tomography import data as ekdata

__cpu_count__ = cpu_count()


class Phase(str, Enum):
    p = 'P'
    s = 'S'

    def __str__(self):
        return self.value

    def __expr__(self):
        return self.__str__()

    def __call__(self):
        return self.__str__()


class EventData(BaseModel):
    location: conlist(float, min_items=3, max_items=3)
    location_correction: conlist(float, min_items=3, max_items=3) = [0, 0, 0]
    resource_id: str=None
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_id = ResourceIdentifier()

    @property
    def loc(self):
        return np.array(self.location)

    @property
    def id(self):
        return self.resource_id.id

    def __str__(self):
        return f'{self.id},{self.location},{self.location_correction}'

    def __repr__(self):
        return str(self)
    

class EventEnsemble(object):
    def __init__(self, events: List[EventData] = []):
        self.events = events
        self.dict = {}
        self.ids = np.arange(0, len(events))
        for event in events:
            self.dict[event.id] = event
            
    def __len__(self):
        return len(self.events)
    
    def __getitem__(self, items):
        if isinstance(items, (list, np.ndarray)):
            events = []
            for item in items:
                if isinstance(item, str):
                    for event in self.events:
                        if event.id == item:
                            events.append(event)
                elif isinstance(item, int):
                    events.append(self.events[item])
                else:
                    raise TypeError
            return EventEnsemble(events=events)
        elif isinstance(items, str):
            for event in self.events:
                if event.id == items:
                    return EventEnsemble(events=[event])
        elif isinstance(items, int):
            return EventEnsemble(events=[self.events[items]])
        else:
            raise TypeError

    def __str__(self):
        out_str = ''
        for event in self.events:
            out_str += f'{event}\n'
        return out_str

    def __repr__(self):
        return str(self)

    def select(self, ids: Union[List[int], int]):
        return self[ids]
    
    def append(self, event: EventData):
        self.events.append(event)
        self.dict[event.id] = event

    @property
    def locs(self):
        locs = []
        for event in self.events:
            locs.append(event.location)
        return locs

    def to_ek_event_table_data(self):
        ev_data = [(event.resource_id, event_id, event.location,
                 event.location_correction) for event, event_id
                in zip(self.events, self.ids)]:
         data.EKEventTable()


class ArrivalTimeData(BaseModel):
    event_id: str
    site_id: str
    phase: Phase
    resource_id: str=None
    arrival_time: datetime

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_id = ResourceIdentifier()


    @property
    def id(self):
        return self.resource_id.id

    def __add__(self, other):

        if not isinstance(other, type(self)):
            raise TypeError(f'operation not permitted between {type(self)} '
                            f'and {type(other)}')

        if self.event_id != other.event_id:
            raise ValueError(f'operation only permitted between two objects'
                             f'with the same event id')

        return self.arrival_time + other.arrival_time

    def __repr__(self):
        return f'event id: {self.event_id}, site id: {self.site_id}, ' \
               f'arrival time:{self.arrival_time}'


class ArrivalTimeEnsemble():
    def __init__(self, arrival_times: List[ArrivalTimeData] = []):
        self.arrival_times = arrival_times

    def append(self, arrival_time):
        if not isinstance(arrival_time, ArrivalTimeData):
            raise TypeError
        self.arrival_times.append(arrival_time)

    def __repr__(self):
        out_str = ''
        for arrival_time in self.arrival_times:
             out_str += f'{str(arrival_time)}\n'

        return out_str

    def __getitem__(self, items):
        if isinstance(items, (list, np.ndarray)):
            arrival_times = []
            for arrival_time in self.arrival_times:
                if arrival_time.id in list(items):
                    arrival_times.append(arrival_time)
            return ArrivalTimeEnsemble(arrival_times=arrival_times)

        elif isinstance(items, str):
            for arrival_time in self.arrival_times:
                if arrival_time.id == items:
                    return arrival_time

        else:
            raise TypeError

    def to_dict(self):
        out_dict = {'id': [],
                    'event_id': [],
                    'site_id': [],
                    'arrival_time': [],
                    'phase': [],
                    'resource_id': []}

        for arrival_time in self.arrival_times:
            out_dict['id'].append(arrival_time.id)
            out_dict['event_id'].append(arrival_time.event_id)
            out_dict['site_id'].append(arrival_time.site_id)
            out_dict['arrival_time'].append(arrival_time.arrival_time)
            out_dict['phase'].append(arrival_time.phase)
            out_dict['resource_id'].append(arrival_time.resource_id)

        return out_dict

    @property
    def event_ids(self):
        event_ids = [arrival_time.event_id for arrival_time
                     in self.arrival_times]

        return list(np.unique(event_ids))

    @property
    def site_ids(self):
        site_ids = [arrival_time.site_id for arrival_time
                     in self.arrival_times]

        return list(np.unique(site_ids))

    def groupby(self, column, **kwargs):
        df = pd.DataFrame(self.to_dict()).set_index('id')
        return df.groupby(column, **kwargs)


class TomoRay(Ray):
    def __init__(self, ray: Ray):
        self.__dict__ = {}
        for key in ray.__dict__.keys():
            self.__dict__[key] = ray.__dict__[key]

        dist = np.array([0] + list(np.linalg.norm(np.diff(self.nodes, 
                                                          axis=0),axis=1)))
        self.dist = np.cumsum(dist)
        
        self.int_x = sc.interpolate.interp1d(self.dist, self.nodes[:, 0])
        self.int_y = sc.interpolate.interp1d(self.dist, self.nodes[:, 1])
        self.int_z = sc.interpolate.interp1d(self.dist, self.nodes[:, 2])
        
        self.velocity = None
        self.epsilon = 10
        self.threshold = 0.1
        self.sensitivity_kernel = None
        
    def add_velocity(self, velocity: VelocityGrid3D):
        self.velocity = velocity
        x, y, z, v = velocity.flattens()
        self.vel_x = x
        self.vel_y = y
        self.vel_z = z
        self.flatten_vel = v
        
    def set_epsilon(self, epsilon):
        self.epsilon

    def sensitivity(self, distance_along_ray):
        """
        measures the sensitivity with respect to the velocity model parameters

        :NOTE: currently, the sensitivity assumes a rbf inverse distance
        interpolation kernel
        """

        from time import time
        t0 = time()
        location = self.interpolate_coordinate(distance_along_ray)

        result = self.rbf_interpolation_sensitivity(location)

        # result = self.velocity.rbf_interpolation_sensitivity(location,
        #                                                      self.epsilon,
        #                                                      threshold=
        #                                                      self.threshold)
        t1 = time()
        logger.info(f'dones measuring sensitivity in {t1 - t0}')

        return result

        # return result.reshape(np.prod(result.shape))
    
    def interpolate_coordinate(self, distance_along_ray):
        """
        return the coordinates at a given distance along the ray
        """

        if not (0 <= distance_along_ray <= np.max(self.dist)):
            raise ValueError(f'the value for "distance_along_ray" must be'
                             f'larger or equal to 0 and smaller or equal to '
                             f'{np.max(self.dist)}')

        return np.array([self.int_x(distance_along_ray),
                         self.int_y(distance_along_ray),
                         self.int_z(distance_along_ray)])

    def rbf_interpolation_sensitivity(self, location):
        """
        calculate the sensitivity of each element given a location
        :param location: location in model space at which the interpolation
        occurs
        :param epsilon: the standard deviation of the gaussian kernel
        :param threshold: threshold relative to the maximum value below which
        the weights are considered 0.
        :rparam: the sensitivity matrix
        """

        # calculating the distance between the location and every grid points

        dist = np.sqrt((self.vel_x - location[0]) ** 2 +
                       (self.vel_y - location[1]) ** 2 +
                       (self.vel_z - location[2]) ** 2)

        sensitivity = np.exp(-(dist / self.epsilon) ** 2)
        sensitivity[sensitivity < np.max(sensitivity) * self.threshold] = 0
        # sensitivity = sensitivity / np.sum(sensitivity)

        return sensitivity
    
    def integrate_sensitivity(self, velocity: VelocityGrid3D, epsilon=10,
                              threshold=0.1, max_it=100, compress_matrix=True):
        from time import time
        t0 = time()
        self.set_epsilon(epsilon)
        self.add_velocity(velocity)
        self.threshold = threshold

        y0 = np.zeros(np.prod(velocity.shape))
        slowness = 1 / velocity.data.reshape(y0.shape)
        sensitivity = y0
        for i, dl in enumerate(np.diff(self.dist)):
            distance_along_ray = self.dist[i] + dl
            sensitivity += self.sensitivity(distance_along_ray) * slowness * dl

        # sensitivity = sensitivity / np.sum(sensitivity) * self.travel_time

        if compress_matrix:
            sensitivity = csr_matrix(sensitivity)

        self.sensitivity_kernel = sensitivity
        t1 = time()
        logger.info(f'done calculating sensitivity in {t1 - t0:0.2f} seconds')

        return

        # to calculate frechet just making sure the sensitivity is normalized,
        # the sensitivity is expressed in s (time)


        # integrator = sc.integrate.RK45(self.sensitivity, 0, y0, self.length)
        #
        # for i in range(0, max_it):
        #     integrator.step()
        #     if integrator.status == 'finished':
        #         break
        #
        # return integrator.y / np.sum(integrator.y)


class TomoRayEnsemble(BaseModel):
    def __init__(self, events: EventEnsemble, arrivals: ArrivalTimeEnsemble,
                 rays: List[Ray] = []):
        self.events = events
        self.arrivals = arrivals
        self.rays = rays

        self.site_ids = []
        for ray in rays:
            self.site_ids.append(ray.site_code)
            # for loc

    def append(self, ray: Ray):
        self.rays.append()


class Tomography(ProjectManager):
        def __init__(self, base_projects_path, project_name, network_code,
                     use_srces=False, solve_velocity=True, solve_location=True,
                     **kwargs):
            
            self.events = None
            self.observations = None
            self.rays = None
            self.solve_location = solve_location
            self.solve_velocity = solve_velocity
            
            super().__init__(base_projects_path, project_name, network_code,
                             use_srces=use_srces, **kwargs)

            self.paths.tomography = self.paths.root / 'tomography'

            super().__init__(base_projects_path, project_name, network_code,
                             use_srces=use_srces, **kwargs)
            
        
        def synthetic(self, dims=[100, 100, 100], 
                      origin=[0, 0, 0], 
                      spacing=[10, 10, 10], 
                      p_mean=5000, p_std=200, s_mean=3000, s_std=125,
                      model_smoothing=3,
                      nsites=20, nevents=1000,
                      trave_time_completeness=0.7, perturbation=0.005,
                      random_seed=None, multi_threaded=True):
            
            self.__add_random_velocities__(dims, origin, spacing, 
                                           p_mean, p_std, s_mean, s_std,
                                           model_smoothing,
                                           random_seed=random_seed)
            self.__add_random_sites__(nsites, random_seed=random_seed)
            self.__add_random_events__(nevents, random_seed=random_seed)
            self.init_travel_time_grids(multi_threaded=multi_threaded)
            self.__add_random_travel_times__()


        def __add_random_velocities__(self, dims, origin, spacing, 
                                      p_mean, p_std, s_mean, s_std,
                                      smoothing, random_seed=None,
                                      multi_threaded=True):
            
            p_velocity = VelocityGrid3D(self.network_code, dims, origin,
                                        spacing)

            s_velocity = VelocityGrid3D(self.network_code, dims, origin,
                                        spacing, phase='S')
            
            p_velocity.fill_random(p_mean, p_std, smoothing, random_seed)

            if random_seed is not None:
                random_seed += 10

            s_velocity.fill_random(s_mean, s_std, smoothing, random_seed)

            # self.add_velocity(p_velocity)
            # self.add_velocity(s_velocity)

            self.add_velocities(VelocityGridEnsemble(p_velocity, s_velocity),
                                initialize_travel_times=False)

        def __add_random_sites__(self, nstations, random_seed=None):
            """
            create random stations
            """

            if not self.has_p_velocity():
                logger.info('The project does not contain a p wave velocity...'
                            ' exiting')
                return

            if not self.has_s_velocity():
                logger.info('The project does not contain a s wave velocity...'
                            ' exiting')
                return

            sta_locs = self.p_velocity.generate_random_points_in_grid(
                nstations, seed=random_seed)

            sites = [Site(label=f'STA{i:02d}', x=sta_loc[0], y=sta_loc[1],
                          z=sta_loc[2]) for i, sta_loc in enumerate(sta_locs)]
            
            self.add_srces(Srces(sites), initialize_travel_time=False)
            
        def __add_random_events__(self, nevents, random_seed=None):
            if not self.has_p_velocity():
                logger.info('The project does not contain a p wave velocity...'
                            ' exiting')
                return

            if not self.has_s_velocity():
                logger.info('The project does not contain a s wave velocity...'
                            ' exiting')
                return
            
            events = self.p_velocity.\
                generate_random_points_in_grid(nevents, seed=random_seed)

            self.events = EventEnsemble()
            for id, event in enumerate(events):
                self.events.append(EventData(location=list(event)))
            
            return self.events
            
        def __add_random_travel_times__(self, completeness=0.6,
                                        min_observations=5,
                                        p_pick_error=0.001,
                                        s_pick_error=0.001):
            """
            generate random travel time observations.
            :param completeness: on average the number of observation per 
            event-sensor pair
            :param min_observations: minimum number of observation for each
            event
            :param p_pick_error: standard deviation of the gaussian
            perturbation in second to add to the travel time for the p picks
            :param s_pick_error: standard deviation of the gaussian
            perturbation in second to add to the travel time for the s picks
            """

            origin_time = UTCDateTime()
            self.arrival_times = ArrivalTimeEnsemble()
            perturbation = {'P': p_pick_error,
                            'S': s_pick_error}
            i = 0
            for event_id in self.events.dict.keys():
                arrivals = []
                for sensor in self.srces:
                    for phase in ['P', 'S']:
                        travel_time = self.travel_times.select(
                            sensor.label, phase=phase)[0].interpolate(
                            self.events.dict[event_id].loc)[0]

                        travel_time += np.random.randn() * perturbation[phase]
                        pick_time = origin_time + travel_time

                        arrival = ArrivalTimeData(event_id=event_id,
                                                  site_id=sensor.label,
                                                  arrival_time=pick_time,
                                                  phase=phase)

                        self.arrival_times.append(arrival)

                        i += 1

        @staticmethod
        def ray_tracer(velocity, data):
            travel_time_grid = data[0]
            arrival_id = data[1]
            loc = data[2]

            ray = travel_time_grid.ray_tracer(loc)
            ray.arrival_id = arrival_id
            # if ray.phase == 'P':
            #     velocity = self.p_velocity
            # else:
            #     velocity = self.s_velocity
            tomo_ray = TomoRay(ray)
            tomo_ray.integrate_sensitivity(velocity)

            return tomo_ray

        def ray_tracing(self, cpu_utilisation=0.9):
            """
            calculate the rays for every station event pair
            """
            num_threads = int(np.ceil(cpu_utilisation * __cpu_count__))

            arrival_time_grouped = self.arrival_times.groupby(
                ['site_id', 'phase'])

            rays = []
            for site_id in tqdm(self.arrival_times.site_ids):
                for phase in Phase:
                    df = arrival_time_grouped.get_group((site_id, phase))

                    tt = self.travel_times.select(site_id,
                                                  phase=phase.value)[0]

                    locs = self.events[df.event_id.values].locs
                    events = self.events[df.event_id.values]
                    arrival_ids = df.resource_id.values
                    tts = [tt] * len(arrival_ids)

                    data = [(travel_time, arrival_id, loc)
                            for travel_time, arrival_id, loc
                            in zip(tts, arrival_ids, locs)]

                    if phase == 'P':
                        velocity = self.p_velocity
                    else:
                        velocity = self.s_velocity

                    ray_tracer = partial(self.ray_tracer, velocity)

                    for d in data:
                        ray_tracer(d)

                    with Pool(num_threads) as pool:
                        rays_tmp = list(tqdm(pool.imap(ray_tracer,
                                                       data),
                                             total=len(locs)))

                        for ray in rays_tmp:
                            rays.append(ray)

                    return rays


                
                



            
            


import pandas
from uquake.core import read, read_events, read_inventory, UTCDateTime
from uquake.core.trace import Trace
from uquake.core.logging import logger
from pathlib import Path
import pandas as pd
from multiprocessing import Pool, cpu_count
import numpy as np
from tqdm import tqdm
import pickle
from useis.core.project_manager import ProjectManager
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
import matplotlib.pyplot as plt
import torch
from uquake.waveform.pick import calculate_snr
from PIL import Image, ImageOps
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from dataset import spectrogram


def get_waveform(waveform_file, inventory):
    st = read(waveform_file)
    for i, tr in enumerate(st):
        for sensor in inventory.sensors:
            if sensor.alternate_code == tr.stats.station:
                st[i].stats.network = inventory[0].code
                st[i].stats.station = sensor.station.code
                st[i].stats.location = sensor.location_code
                for channel in sensor.channels:
                    if tr.stats.channel in channel.code:
                        st[i].stats.channel = channel.code
                        break
                break
    return st


def get_event_id(filename):
    cat = read_events(filename)
    return str(filename.stem), cat[0].resource_id.id


event_type_lookup = {'anthropogenic event': 'noise',
                     'acoustic noise': 'noise',
                     'reservoir loading': 'noise',
                     'road cut': 'noise',
                     'controlled explosion': 'blast',
                     'quarry blast': 'blast',
                     'earthquake': 'seismic event',
                     'sonic boom': 'noise',
                     'collapse': 'noise',
                     'other event': 'noise',
                     'thunder': 'noise',
                     'induced or triggered event': 'noise',
                     'explosion': 'blast',
                     'experimental explosion': 'blast'}

input_data_dir = Path('/data_1/ot-reprocessed-data/')
output_data_dir = Path('/data_1/pick_dataset')
output_data_dir.mkdir(parents=True, exist_ok=True)
sampling_rate = 6000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
number_sample = 256
buffer_image_fraction = 0.05


def create_training_signal(waveform_file):
    # logger.info('here')
    waveform_file_path = input_data_dir / waveform_file
    cat = read_events(waveform_file_path.with_suffix('.xml'))
    st = read(waveform_file_path)

    event_time = cat[0].preferred_origin().time.timestamp

    try:
        st = st.detrend('demean').detrend(
            'linear').taper(max_percentage=0.1,
                            max_length=0.01).resample(
            sampling_rate=sampling_rate)
    except Exception as e:
        logger.error(e)
        trs = []
        for tr in st:
            if np.nan in tr.data:
                continue
            trs.append(tr.copy())
        st.traces = trs

    trs = []
    snrs = []
    pick_times = []
    logger.info(event_type_lookup[cat[0].event_type])
    if event_type_lookup[cat[0].event_type] == 'seismic event':
        for arrival in cat[0].preferred_origin().arrivals:
            site = arrival.pick.site
            for tr in st.select(site=site):
                snr = calculate_snr(tr, arrival.pick.time,
                                    pre_wl=20e-3,
                                    post_wl=20e-3)
                if snr < snr_threshold:
                    continue
                tr.trim(starttime=tr.stats.starttime,
                        endtime=tr.stats.starttime +
                        sequence_length_second,
                        pad=True,
                        fill_value=0)
                trs.append(tr.copy().resample(sampling_rate=
                                              int(sampling_rate)))
                snrs.append(snr)
                pick_times.append(arrival.pick.time)

    else:
        return

    ct = 0
    for tr, pick_time in zip(trs, pick_times):
        for i in range(0, 5):
            try:

                start = np.random.randint(0, number_sample -
                                          int(0.25 * number_sample))

                pick_sample = int((pick_time - tr.stats.starttime) *
                                  tr.stats.sampling_rate)

                start_sample = int(pick_sample - start)

                data = tr.data[start_sample: start_sample + number_sample]
                data = data - np.mean(data)
                data = np.abs(data) / np.abs(data).max()
                pick = start

                out_dict = {'data': data, 'pick': pick}

                out_dir = output_data_dir
                out_dir.mkdir(parents=True, exist_ok=True)
                filename = f'{event_time:0.0f}_{tr.stats.site}_' \
                           f'{tr.stats.channel}_{i}.pickle'

                with open(out_dir / filename, 'wb') as file_out:
                    pickle.dump(out_dict, file_out)
            except Exception as e:
                logger.error(e)


if __name__ == '__main__':
    file_list = [f for f in input_data_dir.glob('*.mseed')]
    for f in tqdm(file_list):
        create_training_signal(f)

    # file_list = [f for f in input_data_dir.glob('*.mseed')]
    # with Pool(4) as pool:
    #     event_list = list(tqdm(pool.map(create_training_image, file_list),
    #                            total=len(file_list)))
    #     # with tqdm(total=len(file_list)) as p_bar:
    #     #     for i, _ in enumerate(pool.imap(create_training_image, file_list)):
    #     #         logger.info('here')
    #     #         p_bar.update()



import pandas
from uquake.core import read, read_events, read_inventory, UTCDateTime, Stream
from uquake.core.trace import Trace
from uquake.core.logging import logger
from pathlib import Path
import pandas as pd
from multiprocessing import Pool, cpu_count
import numpy as np
from tqdm import tqdm
import pickle
from useis.core.project_manager import ProjectManager
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
import matplotlib.pyplot as plt
import torch
from uquake.waveform.pick import calculate_snr
from PIL import Image, ImageOps
from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from dataset import spectrogram


from tqdm.contrib.concurrent import process_map


def get_waveform(waveform_file, inventory):
    st = read(waveform_file)
    for i, tr in enumerate(st):
        for sensor in inventory.sensors:
            if sensor.alternate_code == tr.stats.station:
                st[i].stats.network = inventory[0].code
                st[i].stats.station = sensor.station.code
                st[i].stats.location = sensor.location_code
                for channel in sensor.channels:
                    if tr.stats.channel in channel.code:
                        st[i].stats.channel = channel.code
                        break
                break
    return st


def get_event_id(filename):
    cat = read_events(filename)
    return str(filename.stem), cat[0].resource_id.id


event_type_lookup = {'anthropogenic event': 'noise',
                     'acoustic noise': 'noise',
                     'reservoir loading': 'noise',
                     'road cut': 'noise',
                     'controlled explosion': 'blast',
                     'quarry blast': 'blast',
                     'earthquake': 'seismic event',
                     'sonic boom': 'noise',
                     'collapse': 'noise',
                     'other event': 'noise',
                     'thunder': 'noise',
                     'induced or triggered event': 'noise',
                     'explosion': 'blast',
                     'experimental explosion': 'blast'}

input_data_dir = Path('/data_1/ot-reprocessed-data/')
output_data_dir = Path('/data_1/classification_dataset_1D')
output_data_dir.mkdir(parents=True, exist_ok=True)
sampling_rate = 6000
# num_threads = int(np.ceil(cpu_count() - 10))
num_threads = 10
replication_level = 5
snr_threshold = 10
sequence_length_second = 2
perturbation_range_second = 1
image_width = 128
image_height = 128
buffer_image_fraction = 0.05

buffer_image_sample = int(image_width * buffer_image_fraction)

hop_length = int(sequence_length_second * sampling_rate //
                 (image_width + 2 * buffer_image_sample))


def create_training_dataset(waveform_file):
    # logger.info('here')
    waveform_file_path = input_data_dir / waveform_file
    cat = read_events(waveform_file_path.with_suffix('.xml'))
    st = read(waveform_file_path)

    event_time = cat[0].preferred_origin().time.timestamp

    trs = []
    for tr in st.copy():
        if np.any(np.isnan(tr.data)):
            continue
        trs.append(tr.copy())

    st2 = Stream(traces=trs)

    # try:
    st2 = st2.detrend('demean').detrend('linear')
    st2 = st2.taper(max_percentage=0.1,
                  max_length=0.01)
    st2 = st2.resample(sampling_rate=sampling_rate)
    # except Exception as e:
    #     logger.error(e)
    #     trs = []
    #     for tr in st.copy():
    #         try:
    #             tr = tr.detrend('demean').detrend('linear')
    #             tr = tr.taper(max_percentage=0.1,
    #                           max_length=0.01)
    #             tr = tr.resample(sampling_rate=sampling_rate)
    #             trs.append(tr.copy())
    #         except Exception as e:
    #             logger.error(e)
    #             continue
    #
    #     st2 = Stream(traces=trs)

    trs = []
    # logger.info(event_type_lookup[cat[0].event_type])
    if event_type_lookup[cat[0].event_type] == 'seismic event':
        for arrival in cat[0].preferred_origin().arrivals:
            site = arrival.pick.site
            for tr in st2.select(site=site).copy():
                # filtering out low frequency before calculating the SNR
                tr = tr.filter('highpass', freq=100)
                snr = calculate_snr(tr, arrival.pick.time,
                                    pre_wl=20e-3,
                                    post_wl=20e-3)
                if arrival.pick.evaluation_mode == 'automatic':
                    if snr < snr_threshold:
                        continue
                tr.trim(starttime=tr.stats.starttime,
                        endtime=tr.stats.starttime +
                        sequence_length_second,
                        pad=True,
                        fill_value=0)
                trs.append(tr.copy().resample(sampling_rate=
                                              int(sampling_rate)))

    elif event_type_lookup[cat[0].event_type] == 'blast':
        for tr in st2:
            if (np.max(np.abs(tr.data)) / np.std(tr.data)) < 6:
                continue
            trs.append(tr.trim(endtime=tr.stats.starttime +
                               sequence_length_second))

    else:
        for tr in st2:
            trs.append(tr.trim(endtime=tr.stats.starttime +
                               sequence_length_second))

    ct = 0
    for tr in trs:
        event_type = event_type_lookup[cat[0].event_type]
        filename = f'{event_time:0.0f}_{tr.stats.site}_' \
                   f'{tr.stats.channel}.pickle'
        output_dir = output_data_dir / event_type
        output_file = output_dir / filename

        output_dir.mkdir(parents=True, exist_ok=True)

        out_dict = {'data': tr.data,
                    'seismogram file': waveform_file_path,
                    'event type': event_type_lookup[cat[0].event_type]}

        with open(output_file, 'wb') as f_out:
            pickle.dump(out_dict, f_out)


if __name__ == '__main__':
    file_list = [f for f in input_data_dir.glob('*.mseed')]
    with Pool(20) as pool:
        event_list = list(tqdm(pool.imap(create_training_dataset, file_list),
                               total=len(file_list)))
        # with tqdm(total=len(file_list)) as p_bar:
    # process_map(create_training_dataset, file_list, max_worker=10)
    #     # r = list(tqdm(pool.imap(create_training_dataset, file_list),
    #     #          total=len(file_list)))
                # p_bar.update()
    # for f in tqdm(file_list):
    #     create_training_dataset(f)



from fastapi import FastAPI, Depends, File, UploadFile, Query
import uvicorn
from typing import Optional, List
from io import BytesIO
import pickle
from loguru import logger
from useis.processors.nlloc import NLLOC
from uquake.core import read_inventory
from uquake.nlloc.nlloc import Observations
from uquake.grid.nlloc import VelocityGrid3D
import os
import json
from fastapi.security import OAuth2PasswordBearer
import numpy as np
from useis.services.models import nlloc
from useis.services.models import grid
from pathlib import Path

app = FastAPI()
root_dir = Path(os.environ.setdefault('nll_base_path', 'projects/'))

# oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


def deserialize_object(file_obj):
    f_in = BytesIO(file_obj)
    return pickle.load(f_in)


@app.post("/velocity/{project}/{network}/add/3D/", status_code=201)
async def add_velocity_3d(project: str, network: str,
                          velocity: grid.VelocityGrid3D,
                          initialize_travel_times: Optional[bool] = False):

    nll = NLLOC(root_dir, project, network)

    logger.info(f'adding 3D velocity model to the project:{project} and '
                f'network: {network}')
    velocity_grid = velocity.to_uquake(network)
    return nll.add_velocity(velocity_grid,
                            initialize_travel_times=initialize_travel_times)


@app.post("/inventory/{project}/{network}/simple", status_code=201)
async def add_simple_inventory(project: str, network: str, sites: nlloc.Srces):
    """
       Add or update the inventory using a simple site description

       - **sites**: a list of sites with coordinates and labels
       """
    return


@app.post("/inventory/{project}/{network}", status_code=201)
async def add_inventory(project: str, network: str,
                        inventory: bytes = File(...),
                        initialize_travel_times: Optional[bool] = False):
    nll = NLLOC(root_dir, project, network)
    f_in = BytesIO(inventory)
    inv = read_inventory(f_in)
    logger.info(f'adding inventory to {project}/{network}')
    nll.add_inventory(inv, initialize_travel_time=initialize_travel_times)
    return


@app.post("/srces/{project}/{network}", status_code=201)
async def add_srces(project: str, network: str, srces: bytes = File(...),
                    initialize_travel_times: Optional[bool] = False):
    nll = NLLOC(root_dir, project, network)
    sensors = deserialize_object(srces)
    logger.info(f'adding srces to {project}/{network}')
    nll.add_srces(sensors, force=True,
                  initialize_travel_time=initialize_travel_times)
    return


@app.post('/event/locate/{project}/{network}/',
          response_model=nlloc.NLLOCResults)
async def event_location(project: str, network: str,
                         observations: nlloc.Observations):
    nll = NLLOC(root_dir, project, network)
    event = nll.run_location(observations.to_uquake(), calculate_rays=False)
    print(event.loc)
    return 'nanana'


@app.get('/travel_times/{project}/{network}/init')
async def init_travel_times(project: str, network: str,
                            multi_threaded: Optional[bool] = False):
    nll = NLLOC(root_dir, project, network)
    if multi_threaded:
        thread = 'multi thread'
    else:
        thread = 'single thread'
    logger.info(f'calculating the travel time using {thread}')
    nll.init_travel_time_grids(multi_threaded=multi_threaded)
    return


@app.get("/velocity/{project}/{network}")
async def list_velocity(project, network):
    return project, network


@app.get("/test/{project}/{network}/random_locations")
async def generate_random_locations(project: str, network: str,
                                    n_points: Optional[int] = 1):
    """
    Generate random location(s) in grid

    - **project**: project name
    - **network**: network id
    - **n_points**:  number of points random location to return (Default 1)

    the number of points is controlled by the n_points parameter which
    is set to 1 by default.
    """
    nll = NLLOC(root_dir, project, network)
    logger.info(nll.p_velocity.generate_random_points_in_grid())
    logger.info(n_points)
    observation = nll.p_velocity.generate_random_points_in_grid(
        n_points=n_points)
    return json.dumps(list(observation.ravel()))


@app.get("/test/{project}/{network}/observations",
         response_model=nlloc.Observations)
async def generate_random_observations(project: str, network: str,
                                       x: float, y: float, z: float) \
        -> nlloc.Observations:
    """
    Generate random observation in grid

    - **project**: project name
    - **network**: network id
    - **x**: X coordinates at which to calculate the observations
    - **y**: Y coordinates at which to calculate the observations
    - **z**: Z coordinates at which to calculate the observations
    """
    nll = NLLOC(root_dir, project, network)
    e_loc = np.array([x, y, z])
    print(e_loc)
    observations = Observations.generate_observations_event_location(
        nll.travel_times, e_loc=e_loc)

    print(nlloc.Observations.from_uquake(observations))

    return nlloc.Observations.from_uquake(observations)


@app.get('/projects')
async def get_project_list() -> List[str]:
    projects = [str(f).split(os.path.sep)[-1]
                for f in root_dir.glob('*') if f.is_dir()]
    return projects


@app.get('/network/{project}')
async def get_network_list(project: str) -> List[str]:
    networks = [str(f).split(os.path.sep)[-1]
                for f in root_dir.glob(f'{project}/*') if f.is_dir()]
    return networks

def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="GRID API",
        version="0.1.0",
        description="Your API description",
        routes=app.routes,
    )
    app.openapi_schema = openapi_schema
    return app.openapi_schema

@app.get("/docs", include_in_schema=False)
async def get_documentation():
    return get_swagger_ui_html(
        openapi_url="/openapi.json",
        title="Your API title",
    )

@app.get("/openapi.json", include_in_schema=False)
async def get_openapi_schema():
    return custom_openapi()

def start():
    uvicorn.run("useis.services.grid.server:app", reload=True)







import dataset
from torch.utils.data import DataLoader
import torch.nn as nn
import torch
import numpy as np
from importlib import reload
reload(dataset)
from tqdm import tqdm
import model
import pickle
import matplotlib.pyplot as plt
from loguru import logger
from sklearn.metrics import confusion_matrix
from plotcm import plot_confusion_matrix
import model
reload(model)

input_directory = '/data_1/classification_dataset_2/'
suffix = ''
extension = 'jpg'

# file_list = glob(os.path.join(input_directory, '*', f'*{suffix}.{extension}'))

# training, test = spectrogram_dataset.split_dataset(input_directory, split=0.8,
#                                                    seed=1)

file_list = dataset.FileList(input_directory)

ec = model.EventClassifier(len(file_list.category_list))

losses = []
accuracies = []

batch_size = 1000
for epoch in tqdm(range(0, 100)):
    logger.info('selecting data')
    training_dataset = file_list.select(2e5)
    test_dataset = file_list.select(1e4)
    ec.train(training_dataset, batch_size=batch_size)
    accuracy = ec.validate(test_dataset, batch_size=batch_size)

    logger.info(f'Average iteration loss: {np.mean(ec.losses): 0.3f} '
                f'+/- {np.std(ec.losses): 0.3f}')
    logger.info(f'Average iteration accuracy: {np.mean(ec.accuracies):0.3f} '
                f'+/- {np.std(ec.accuracies): 0.3f}')

    losses.append(np.mean(ec.losses))
    accuracies.append(accuracy)
    # accuracies.append(np.mean(ec.accuracies))

    if epoch % 10 == 0:
        # ec.model.eval()
        ec.save(f'classifier_model_{epoch + 1}_epoch.pickle')

    plt.figure(1)
    plt.clf()
    plt.ion()
    # plt.plot(np.arange(epoch + 1), losses)
    plt.plot(np.arange(epoch + 1), accuracies)
    plt.xlabel('epoch')
    plt.ylabel('loss/accuracy')
    plt.pause(0.1)
    plt.show()




# # pickle.dump(training_dataset, open('training_dataset.pickle', 'wb'))
# # training_dataset = pickle.load(open('training_dataset.pickle', 'rb'))
# # training_dataset = pickle.load(open('training_dataset.pickle', 'rb'))
#
# # test_dataset = spectrogram_dataset.SpectrogramDataset(input_directory,
# #                                                       max_image_sample=4e4)
# # pickle.dump(test_dataset, open('test_dataset.pickle', 'wb'))
# # test_dataset = pickle.load(open('test_dataset.pickle', 'rb'))
#
# nb_pixel = training_dataset.nb_pixel
# nb_categories = training_dataset.nb_categories
#
#
#
# # model = model.CNN(nb_categories)
# # VGG16
# # model = models.vgg16(pretrained=False, progress=True)
# # first_conv_layer = [nn.Conv2d(1, 3, kernel_size=3, stride=1, padding=1,
# #                                dilation=1, groups=1, bias=True)]
# # first_conv_layer.extend(list(model.features))
# # model.features = nn.Sequential(*first_conv_layer)
#
# # resnet 50
# # model = models.resnet34(pretrained=False)
# # weights = model.conv1.weight.clone()
# # model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
# # model.conv1.weight[:, :1] = weights
# # model.conv1.weight[:, 3] = model.conv1.weight[:, 0]
#
#
# # model = nn.Sequential(
# #     nn.Linear(nb_pixel, 128),
# #     nn.ReLU(),
# #     nn.Linear(128, nb_categories)
# # )
#
#
# batch_size = 2000
#
# training_loader = DataLoader(training_dataset, batch_size=batch_size,
#                              shuffle=True)
#
# n_epochs = 25
#
# # Stuff to store
# train_losses = []  # np.zeros(n_epochs)
# train_accuracies = []
#
# # plt.ion()
# # fig, ax = plt.subplots()
# # xlim = 1000
# # # plt.xlim([0, xlim])
# # plt.ylim([0, 1])
# # plt.axhline(0.95, color='k', ls='--')
# # plt.axhline(0.98, color='k', ls='--')
# #
# # fig2, ax2 = plt.subplots()
# # plt.xlim([0, 2300])
# # plt.ylim([0, 2])
# # plt.show()
#
# for it in tqdm(range(n_epochs)):
#
#     for inputs, targets in tqdm(training_loader):
#         inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1],
#                              inputs.size()[2])
#         inputs = inputs.to(device)
#         targets = targets.to(device)
#
#         optimizer.zero_grad()
#
#         # Forward pass
#         outputs = model(inputs)
#         loss = criterion(outputs, targets)
#         tmp_choices = [np.argmax(out.cpu().detach().numpy())
#                        for out in outputs]
#         total_accuracy = np.sum(tmp_choices ==
#                                 targets.cpu().detach().numpy()) \
#                          / len(tmp_choices)
#
#         train_accuracy.append(total_accuracy)
#
#         # Backward and optimize
#         loss.backward()
#         optimizer.step()
#
#         # print(loss.item())
#         train_loss.append(loss.item())
#         iterations.append(i)
#
#         if i % 20 == 19:
#             logger.info(f'accuracy: '
#                         f'{np.mean(np.array(train_accuracy)[-100:-1]): 0.3f} '
#                         f'+/- '
#                         f'{np.std(np.array(train_accuracy)[-100:-1]):0.3f}')
#             cm = confusion_matrix(targets.cpu().detach().numpy(),
#                                   np.array(tmp_choices))
#
#             print(cm)
#
#             plt.figure(10)
#             plt.clf()
#             plot_confusion_matrix(cm, training_dataset.category_list,
#                                   normalize=True)
#             plt.show()
#
#
#             # test_loader = DataLoader(test_dataset, batch_size=1000,
#             #                          shuffle=True)
#             #
#             # test_accuracies = []
#             # for t_inputs, t_targets in tqdm(test_loader):
#             #     model = model.to(device)
#             #     inps = t_inputs.view(t_inputs.size()[0], -1,
#             #                          t_inputs.size()[1],
#             #                          t_inputs.size()[2])
#             #     # inps = inps.to(device)
#             #     # targs = t_targets.to(device)
#             #     targs = t_targets.to(device)
#             #     #
#             #     outs = model(inps.to(device))
#             #
#             #     tmp_choices = [np.argmax(out.cpu().detach().numpy())
#             #                    for out in outs]
#             #     test_accuracies.append(np.sum(tmp_choices ==
#             #                                 targs.cpu().detach().numpy()) \
#             #                            / len(tmp_choices))
#             #
#             # logger.info(f'accuracy: {np.mean(test_accuracies)} +/- '
#             #             f'{np.std(test_accuracies)}')
#
#         # if i == 0:
#         #     sc, = ax.plot(iterations, train_loss)
#         #     plt.show()
#         #     sc2, = ax.plot(iterations, train_accuracy)
#         # # if i == 200:
#         # #     optimizer.param_groups[0]['lr'] = \
#         # #         optimizer.param_groups[0]['lr'] * 10
#         # # if i == 400:
#         # #     optimizer.param_groups[0]['lr'] = \
#         # #         optimizer.param_groups[0]['lr'] * 10
#         # else:
#         #     # plt.figure(1)
#         #     # plt.clf()
#         #     # plt.plot(iterations, train_losses)
#         #     # plt.show()
#         #     # ax.plot(iterations, train_losses)
#         #     if np.max(iterations) > xlim:
#         #         xlim += 1000
#         #         plt.xlim([0, xlim])
#         #     sc.set_data(iterations, train_loss)
#         #     sc2.set_data(iterations, train_accuracy)
#         #
#         #     ax = plt.gca()
#         #
#         #     # recompute the ax.dataLim
#         #     ax.relim()
#         #     # update ax.viewLim using the new dataLim
#         #     ax.autoscale_view()
#         #     plt.draw()
#         #
#         #     fig.canvas.draw_idle()
#         #     # fig2.canvas.draw_idle()
#         #     try:
#         #         fig.canvas.flush_events()
#         #         # fig2.canvas.flush_events()
#         #     except NotImplementedError:
#         #         pass
#         i += 1
#         train_losses.append(train_loss)
#         train_accuracies.append(train_accuracy)
#
#     # to avoid over-fitting the model
#     # training_dataset = spectrogram_dataset.SpectrogramDataset(input_directory,
#     #                                                           max_image_sample
#     #                                                           =2e5)
#     training_dataset = file_list.select(1e5)
#     training_loader = DataLoader(training_dataset, batch_size=batch_size,
#                                  shuffle=True)
#
#     # train_losses.append(np.min(train_loss))
#     # train_accuracy.append(torch.sum())
#     # if it == 0:
#     #     sc2, = ax2.plot(it, train_losses)
#     #     plt.plot()
#     # else:
#     #     sc2.set_data(np.arange(len(train_losses)), train_losses)
#     #     fig.canvas.draw_idle()
#     #     try:
#     #         fig.canvas.flush_events()
#     #     except NotImplementedError:
#     #         pass
#
# pickle.dump(model, open('model.pickle', 'rb'))
#
# # cpu_model = model.cpu()
# # test_loader = DataLoader(test_dataset, batch_size=1000, shuffle=True)
# # i = 0
# # acc = 0
# # for t_inputs, t_targets in tqdm(test_loader):
# #     model = model.to(device)
# #     inps = t_inputs.view(t_inputs.size()[0], -1, t_inputs.size()[1],
# #                          t_inputs.size()[2])
# #     # inps = inps.to(device)
# #     # targs = t_targets.to(device)
# #     targs = t_targets.to(device)
# #     #
# #     outputs = model(inps.to(device))
# #
# #     tmp_choices = [np.argmax(out.cpu().detach().numpy())
# #                    for out in outputs]
# #     total_accuracy = np.sum(tmp_choices ==
# #                             targs.cpu().detach().numpy()) \
# #                      / len(tmp_choices)
# #
# #     acc += total_accuracy
# #     i += 1
#     # input(total_accuracy)
#
#
#



from ..core.project_manager import ProjectManager


class RayTracer(ProjectManager):
    """
    Ray tracing module
    """

    def __init__(self, base_projects_path, project_name, network_code,
                 **kwargs):
        super().__init__(base_projects_path, project_name, network_code,
                         **kwargs)
        pass
    


from loguru import logger

from ..core.project_manager import ProjectManager
from uquake.core.stream import Stream

from pathlib import Path
import os
import shutil
from ..settings.settings import Settings
from time import time
from uquake.core.util import tools

from xseis2 import xspy

import numpy as np

class Beamforming(ProjectManager):
    def __init__(self, base_project_path, project_name, network_code, 
                 use_srces: bool = False):

        super().__init__(base_projects_path=base_project_path,
                         project_name=project_name,
                         network_code=network_code,
                         use_srces=use_srces)

    def locate(self, stream: Stream):

        nthreads = self.settings.beamforming.nthreads

        fixed_wlen_sec = self.settings.beamforming.fixed_wlen_sec
        sample_rate_decimated = self.settings.beamforming.samplerate_decimated
        pair_dist_min = self.settings.beamforming.pair_dist_min
        pair_dist_max = self.settings.beamforming.pair_dist_max
        cc_smooth_length_sec = self.settings.beamforming.cc_smooth_length_sec

        whiten_corner_freqs = np.array(
            self.settings.beamforming.whiten_corner_freqs, dtype=np.float32)

        black_list = self.settings.black_list

        logger.info('reading the hdf5 file')
        sitelocs = self.htt.locations
        ttable = (self.htt.hf["ttp"][:] * sample_rate_decimated).astype(
            np.uint16)
        ttable_s = (self.htt.hf["tts"][:] * sample_rate_decimated).astype(
            np.uint16)
        ngrid = ttable.shape[1]
        ttable_row_ptrs = np.array(
            [row.__array_interface__["data"][0] for row in ttable])
        ttable_row_ptrs_s = np.array(
            [row.__array_interface__["data"][0] for row in ttable_s])

        logger.info('preparing data for beamforming')
        t0 = time()

        for trace in stream.composite():
            station = trace.stats.station
            location = trace.stats.location
            channel = trace.stats.channel
            site = trace.stats.site

            if trace.stats.site not in self.travel_times.seed_labels:
                logger.warning(f'Sensor {station}.{location} not in the '
                               f'H5 travel time file sensor list... removing '
                               f'trace for {station}.{location} and '
                               f'component')
                stream.remove(trace)
            elif np.max(trace.data) == 0:
                logger.warning(f'trace for component {channel} of sensor'
                               f' {station}.{location} contains only zero... '
                               f'removing trace')
                stream.remove(trace)
            elif trace.stats.station in self.settings.black_list:
                logger.warning(f'sensor {station}.{location} is in the black '
                               f'list... removing trace for sensor '
                               f'{station}.{location} and channel {channel}')

                stream.remove(trace)

        stream.resample(sampling_rate=sample_rate_decimated)
        data, sample_rate, ts = stream.as_array(fixed_wlen_sec)
        data = np.nan_to_num(data)
        # decimate_factor = int(sample_rate / sample_rate_decimated)
        # data = tools.decimate(data, sample_rate, decimate_factor)
        channel_map = stream.channel_map().astype(np.uint16)

        ikeep = self.htt.index_sites(stream.unique_sites)
        debug_file = os.path.join(self.paths.debug, "iloc_" + str(ts) +
                                  ".npz")

        t1 = time()
        logger.info(
            "done preparing data for Interloc in %0.3f seconds" % (t1 - t0))

        logger.info("Locating event with Interloc")
        t6 = time()
        logger.info(
            "sample_rate_decimated {}, ngrid {}, nthreads {}, debug {}, "
            "debug_file {}",
            sample_rate_decimated, ngrid, nthreads, self.settings.debug_level,
            debug_file)
        
        from ipdb import set_trace
        set_trace()

        out = xspy.pySearchOnePhase(
            data,
            sample_rate_decimated,
            channel_map,
            sitelocs[ikeep],
            ttable_row_ptrs[ikeep],
            ngrid,
            whiten_corner_freqs,
            pair_dist_min,
            pair_dist_max,
            cc_smooth_length_sec,
            nthreads,
            self.settings.debug_level,
            debug_file
        )

        out_s = xspy.pySearchOnePhase(
            data,
            sample_rate_decimated,
            channel_map,
            sitelocs[ikeep],
            ttable_row_ptrs_s[ikeep],
            ngrid,
            whiten_corner_freqs,
            pair_dist_min,
            pair_dist_max,
            cc_smooth_length_sec,
            nthreads,
            self.settings.debug_level,
            debug_file
        )

        vmax, imax, iot = out
        vmax_s, imax_s, iot_s = out_s

        if vmax_s > vmax:
            vmax = vmax_s
            imax = imax_s
            iot = iot_s
            logger.info('stacking along the s-wave moveout curve yielded '
                        'better result')

        normed_vmax = vmax * fixed_wlen_sec
        lmax = self.htt.icol_to_xyz(imax)
        t7 = time()
        logger.info("Done locating event with Interloc in %0.3f seconds" % (
                t7 - t6))

        t0_epoch = t0.timestamp
        ot_epoch = t0_epoch + iot / sample_rate_decimated

        method = "%s" % ("INTERLOC",)

        logger.info("power: %.3f, ix_grid: %d, ix_ot: %d" % (vmax, imax, iot))
        logger.info("utm_loc: {}", lmax)
        logger.info("=======================================\n")
        logger.info("VMAX over threshold (%.3f)" % (vmax))

        self.response = {'x': lmax[0],
                         'y': lmax[1],
                         'z': lmax[2],
                         'vmax': vmax,
                         'normed_vmax': normed_vmax,
                         'event_time': ot_epoch,
                         'method': method}

        return self.response

        return


        

        
        
# import os
# from datetime import datetime
# from time import time
#
# import numpy as np
# from obspy.core import UTCDateTime
# from obspy.core.util.attribdict import AttribDict
# from xseis2 import xspy
#
# from microquake.helpers.logging import logger
# from microquake.core.event import Origin
# from microquake.core.util import tools
#
# from microquake.helpers.hdf5 import get_ttable_h5
# from microquake.processors.processing_unit import ProcessingUnit
#
#
#
#
# class Processor(ProcessingUnit):
#     @property
#     def module_name(self):
#         return "interloc"
#
#     def initializer(self):
#         self.htt = get_ttable_h5()
#
#     def process(
#         self,
#         **kwargs
#     ):
#         logger.info("pipeline: interloc")
#
#         # TODO: copy not necessary test application is broken
#         stream = kwargs["stream"].copy()
#
#         nthreads = self.params.nthreads
#         fixed_wlen_sec = self.params.fixed_wlen_sec
#         sample_rate_decimated = self.params.samplerate_decimated
#         pair_dist_min = self.params.pair_dist_min
#         pair_dist_max = self.params.pair_dist_max
#         cc_smooth_length_sec = self.params.cc_smooth_length_sec
#
#         whiten_corner_freqs = np.array(self.params.whiten_corner_freqs,
#                                        dtype=np.float32)
#
#         stalocs = self.htt.locations
#         ttable = (self.htt.hf["ttp"][:] * sample_rate_decimated).astype(
#             np.uint16)
#         ttable_s = (self.htt.hf["tts"][:] * sample_rate_decimated).astype(
#             np.uint16)
#         ngrid = ttable.shape[1]
#         ttable_row_ptrs = np.array(
#             [row.__array_interface__["data"][0] for row in ttable])
#         ttable_row_ptrs_s = np.array(
#             [row.__array_interface__["data"][0] for row in ttable_s])
#
#         logger.info("preparing data for Interloc")
#         t4 = time()
#
#         # remove channels which do not have matching ttable entries
#         # This should be handled upstream
#
#         for trace in stream:
#             station = trace.stats.station
#             component = trace.stats.channel
#             if trace.stats.station not in self.htt.stations:
#                 logger.warning(f'Sensor {station} not in the H5 travel time '
#                                f'file sensor list... removing trace for '
#                                f'{station} and component {component}')
#                 stream.remove(trace)
#             elif np.max(trace.data) == 0:
#                 # from ipdb import set_trace; set_trace()
#                 logger.warning(f'trace for component {component} of sensor'
#                                f' {station} contains only zero... removing '
#                                f'trace')
#                 stream.remove(trace)
#             elif trace.stats.station in self.settings.sensors.black_list:
#                 logger.warning(f'sensor {station} is in the black list... '
#                                f'removing trace for sensor {station} and '
#                                f'component {component}')
#
#                 stream.remove(trace)
#
#         data, sample_rate, t0 = stream.as_array(fixed_wlen_sec)
#         data = np.nan_to_num(data)
#         decimate_factor = int(sample_rate / sample_rate_decimated)
#         data = tools.decimate(data, sample_rate, decimate_factor)
#         channel_map = stream.channel_map().astype(np.uint16)
#
#         ikeep = self.htt.index_sta(stream.unique_stations())
#         debug_file = os.path.join(self.debug_file_dir, "iloc_" + str(t0) +
#                                   ".npz")
#         t5 = time()
#         logger.info(
#             "done preparing data for Interloc in %0.3f seconds" % (t5 - t4))
#
#         logger.info("Locating event with Interloc")
#         t6 = time()
#         logger.info(
#             "sample_rate_decimated {}, ngrid {}, nthreads {}, debug {}, "
#             "debug_file {}",
#             sample_rate_decimated, ngrid, nthreads, self.debug_level,
#             debug_file)
#
#         out = xspy.pySearchOnePhase(
#             data,
#             sample_rate_decimated,
#             channel_map,
#             stalocs[ikeep],
#             ttable_row_ptrs[ikeep],
#             ngrid,
#             whiten_corner_freqs,
#             pair_dist_min,
#             pair_dist_max,
#             cc_smooth_length_sec,
#             nthreads,
#             self.debug_level,
#             debug_file
#         )
#
#         out_s = xspy.pySearchOnePhase(
#             data,
#             sample_rate_decimated,
#             channel_map,
#             stalocs[ikeep],
#             ttable_row_ptrs_s[ikeep],
#             ngrid,
#             whiten_corner_freqs,
#             pair_dist_min,
#             pair_dist_max,
#             cc_smooth_length_sec,
#             nthreads,
#             self.debug_level,
#             debug_file
#         )
#
#         vmax, imax, iot = out
#         vmax_s, imax_s, iot_s = out_s
#
#         if vmax_s > vmax:
#             vmax = vmax_s
#             imax = imax_s
#             iot = iot_s
#             logger.info('stacking along the s-wave moveout curve yielded '
#                         'better result')
#
#         normed_vmax = vmax * fixed_wlen_sec
#         lmax = self.htt.icol_to_xyz(imax)
#         t7 = time()
#         logger.info("Done locating event with Interloc in %0.3f seconds" % (
#                 t7 - t6))
#
#         t0_epoch = t0.timestamp
#         ot_epoch = t0_epoch + iot / sample_rate_decimated
#
#         method = "%s" % ("INTERLOC",)
#
#         logger.info("power: %.3f, ix_grid: %d, ix_ot: %d" % (vmax, imax, iot))
#         logger.info("utm_loc: {}", lmax)
#         logger.info("=======================================\n")
#         logger.info("VMAX over threshold (%.3f)" % (vmax))
#
#         self.response = {'x': lmax[0],
#                          'y': lmax[1],
#                          'z': lmax[2],
#                          'vmax': vmax,
#                          'normed_vmax': normed_vmax,
#                          'event_time': ot_epoch,
#                          'method': method}
#
#         return self.response
#
#     def output_catalog(self, catalog):
#         catalog = catalog.copy()
#
#         x = self.response['x']
#         y = self.response['y']
#         z = self.response['z']
#         vmax = self.response['vmax']
#         normed_vmax = self.response['normed_vmax']
#         method = self.response['method']
#         event_time = UTCDateTime(datetime.utcfromtimestamp(self.response[
#                                                             'event_time']))
#
#         catalog[0].origins.append(
#             Origin(x=x, y=y, z=z, time=event_time,
#                    method_id=method, evalution_status="preliminary",
#                    evaluation_mode="automatic")
#         )
#         catalog[0].preferred_origin_id = catalog[0].origins[-1].resource_id.id
#
#         catalog[0].preferred_origin().extra.interloc_vmax \
#             = AttribDict({'value': vmax, 'namespace': 'MICROQUAKE'})
#
#         catalog[0].preferred_origin().extra.interloc_normed_vmax \
#             = AttribDict({'value': normed_vmax, 'namespace': 'MICROQUAKE'})
#
#         return catalog
#
#     def legacy_pipeline_handler(
#         self,
#         msg_in,
#         res
#     ):
#         cat, stream = self.app.deserialise_message(msg_in)
#
#         cat = self.output_catalog(cat)
#
#         return cat, stream



from useis.processors import classifier
from importlib import reload
from useis.ai import model
from uquake.core.logging import logger
from tqdm import tqdm
import numpy as np
import torch
from torchvision import models
import matplotlib.pyplot as plt
from uquake import read
reload(classifier)
reload(model)

torch.cuda.empty_cache()

classifier_project = classifier.Classifier('/data_1/projects/', 'classifier', 'OT',
                                           reset_training=False)

classifier_project.train(override_current_model=True)
# classifier_project.train()

# train, test, validation = classifier_project.split_dataset(use_synthetic=False)

torch.save(classifier_project.event_classifier.model.state_dict,
           f'/data_1/ai_models/{datetime.now()}_classification_model.pth')


ec = classifier_project.event_classifier

out_dict = {'model_state': ec.model.state_dict(),
            'out_features': ec.n_features,
            'label_map': ec.label_mapping,
            'learning_rate': ec.learning_rate,
            'model_id': ec.model_id}

pickle.dump(out_dict, open('/data_1/ai_models/test.pickle', 'wb'))

def read_model(input_file):
    in_dict = pickle.load(open('/data_1/ai_models/test.pickle', 'rb'))

    # def __init__(self, n_features: int, label_mapping, gpu: bool = True,
    #              learning_rate: float = 0.001,
    #              model=models.resnet34(), model_id=None):
    ec2 = model.EventClassifier(in_dict['out_features'], in_dict['label_map'], gpu=False,
                                learning_rate=in_dict['learning_rate'],
                                model_id=in_dict['model_id'])
    ec2.model.load_state_dict(in_dict['model_state'])
    ec2.model.eval()



classifier_project



from useis.processors import classifier
from importlib import reload
from useis.ai import model
from uquake.core.logging import logger
from tqdm import tqdm
import numpy as np
import torch
from torchvision import models
import matplotlib.pyplot as plt
from uquake import read, read_events
from pathlib import Path
from uquake.core.stream import Stream

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import numpy as np
import uquake
reload(classifier)
reload(model)

torch.cuda.empty_cache()

classifier_project = classifier.Classifier('/data_1/projects/', 'classifier', 'OT',
                                           reset_training=False)

# classifier_model = model.EventClassifier.load()
# classifier_project.add_model(classifier_model)

records = classifier_project.training_db_manager.filter(categories='blast')
# classifier_project.train()


def plot_results(stream, labels, attention_end):

    # Assuming `streams` is a list of obspy streams and `labels` is a list of labels
    fig, axs = plt.subplots(nrows=len(stream), sharex=True, sharey=True,
                            figsize=(15, 12))

    attention_end = attention_end - stream[0].stats.starttime
    attention_start = attention_end - 1

    # Iterate over the streams and plot each one
    for i, (trtr, label) in enumerate(zip(stream, labels)):
        try:
            axes = axs[i]
        except Exception as e:
            axes = axs
        t = np.arange(len(trtr.data)) / trtr.stats.sampling_rate
        trtr.data = trtr.data / trtr.data.max()
        axes.plot(t, trtr.data, 'k')

        # Add labels and title to the plot
        axes.set_ylabel(f"Trace {i + 1}")
        # axes.set_xlabel("Time (s)")
        # axes.set_title(f"{label}\nTrace {i + 1}")

        axes.set_ylim([-1.1, 1.1])

        # Draw a rectangle to identify the attention window
        # attention_start = trace.stats.starttime + 0.5
        # attention_end = attention_start + 1.0
        rect = Rectangle((attention_start, axes.get_ylim()[0]),
                         attention_end - attention_start,
                         axes.get_ylim()[1] - axes.get_ylim()[0],
                         linewidth=1, edgecolor='r', facecolor='r', alpha=0.2)
        axes.add_patch(rect)

        axes.text(0.95, 0.95, label, transform=axes.transAxes,
                    fontsize=10, ha='right', va='top',
                    bbox=dict(facecolor='none', edgecolor='none'))

    # Add a common title for the figure
    # plt.subtitle("Multiple Time Series with Attention Windows")
    try:
        axs[-1].set_xlabel("Time (s)")
    except:
        axs.set_xlabel("Time (s)")

    # Adjust the layout and show the plot
    # fig.subplots_adjust(hspace=0)
    fig.tight_layout()
    # plt.xlim([0, 2])
    plt.subplots_adjust(hspace=0)
    plt.show()


root_dir = Path('/data_1/ot-reprocessed-data/')

filenames = np.unique([record.mseed_file for record in records])

for context_mseed in filenames[121:]:
    # input(context_mseed)
    catalog_file = root_dir / context_mseed.replace('context_mseed', 'xml')
    mseed_file = root_dir / context_mseed.replace('context_mseed', 'mseed')

    cat = read_events(str(catalog_file))
    st = read(str(mseed_file)).select(channel='*Z')
    # st = read(str(mseed_file)).composite()
    st_context = read(root_dir / context_mseed)

    event_loc = cat[0].preferred_origin().loc

    distances = []
    stations = []
    locations = []
    for tr in st:
        sites = classifier_project.inventory.select(station=tr.stats.station,
                                                    location=tr.stats.location).sites
        if len(sites) == 0:
            continue
        site = sites[0]
        distances.append(np.linalg.norm(site.loc - event_loc))
        stations.append(site.station_code)
        locations.append(site.location_code)

    indices = np.argsort(distances)
    distances = np.array(distances)
    stations = np.array(stations)
    locations = np.array(locations)

    sta_loc = [f'{sta}.{loc}' for sta, loc in zip(stations, locations)]

    _, i1 = np.unique(sta_loc, return_index=True)

    indices = indices[i1]

    traces = []
    for index in indices[:10]:
        for tr in st.select(station=stations[index], location=locations[index]).copy():
            # tr.stats.station = f'{int(distances[index]):04d}'
            traces.append(tr)

    if len(traces) == 0:
        continue

    st2 = Stream(traces=traces).detrend('demean').detrend('linear')

    end_time = np.min([tr.stats.starttime for tr in st2]) + 2
    outputs = []
    for end_time_perturbation in [0, 0.5, 1]:
        st3 = st2.copy()
        st3.trim(endtime=end_time - end_time_perturbation).detrend('demean').detrend(
            'linear').taper(max_percentage=0.05)
        output = classifier_project.predict(st3)
        outputs.append(output)
        input(output.predicted_class_ensemble(cat[0].preferred_origin().loc))

        plot_results(st2, output.predicted_classes,
                     attention_end=end_time - end_time_perturbation)

    # end_time = st_context[0].stats.endtime
    # outputs = []
    # for end_time_perturbation in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
    #     st3 = st_context.copy()
    #     st3.trim(endtime=end_time - end_time_perturbation)
    #     output = classifier_project.predict(st3)
    #     outputs.append(output)
    #     print(f'{output.predicted_class_ensemble(cat[0].preferred_origin().loc)}\n'
    #           f'{output.raw_output}\n')
    #
    #     plot_results(st_context, output.predicted_classes,
    #                  attention_end=end_time - end_time_perturbation)




    # # find the traces from the 10 nearest sensors
    # distances = {}
    # for arrival in cat[0].preferred_origin().arrivals:
    #     station_code = arrival.pick.waveform_id.station_code
    #     if station_code in distances.keys():
    #         continue
    #     distances[station_code] = arrival.distance

    # output = classifier_project.predict(st[0:50])

    # krapout




# st = read('/data_1/ot-reprocessed-data/177f51f6d4485d2ab4396ec7b02799f8.mseed')



